import unittest
from testfixtures import TempDirectory

import os.path

from tests.helpers.base_test_case import BaseTestCase

from ploigos_step_runner.config import Config, StepConfig, SubStepConfig, ConfigValue

class TestSubStepConfig(BaseTestCase):
    def test_constructor_no_sub_step_config_or_step_env_config(self):
        sub_step_config = SubStepConfig(
            parent_step_config=None,
            sub_step_name='sub-step-foo',
            sub_step_implementer_name='foo'
        )

        self.assertEqual(sub_step_config.sub_step_config, {})
        self.assertEqual(sub_step_config.sub_step_config, {})

    def test_parent_config(self):
        config = Config({
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'test1': 'foo'
                        }
                    }
                ]

            }
        })

        step_config = config.get_step_config('step-foo')
        sub_step = step_config.get_sub_step('foo1')

        self.assertEqual(sub_step.parent_config, config)

    def test_step_name(self):
        config = Config({
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'test1': 'foo'
                        }
                    }
                ]

            }
        })

        step_config = config.get_step_config('step-foo')
        sub_step = step_config.get_sub_step('foo1')

        self.assertEqual(sub_step.step_name, 'step-foo')

    def test_global_defaults(self):
        config = Config({
            Config.CONFIG_KEY: {
                'global-defaults': {
                    'test1': 'global-default-1',
                    'test2': 'global-default-2'
                },
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'test1': 'foo'
                        }
                    }
                ]

            }
        })

        step_config = config.get_step_config('step-foo')
        sub_step = step_config.get_sub_step('foo1')

        self.assertEqual(sub_step.global_defaults, {
            'test1': ConfigValue(
                'global-default-1',
                None,
                ["step-runner-config", "global-defaults", "test1"]
            ),
            'test2': ConfigValue(
                'global-default-2',
                None,
                ["step-runner-config", "global-defaults", "test2"]
            )
        })

    def test_global_environment_defaults(self):
        config = Config({
            Config.CONFIG_KEY: {
                'global-environment-defaults': {
                    'env1': {
                        'test1': 'global-env1-default-1',
                        'test2': 'global-env1-default-2'
                    },
                    'env2': {
                        'test1': 'global-env2-default-1',
                        'test2': 'global-env2-default-2'
                    }
                },
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'test1': 'foo'
                        }
                    }
                ]

            }
        })

        step_config = config.get_step_config('step-foo')
        sub_step = step_config.get_sub_step('foo1')

        self.assertEqual(
            ConfigValue.convert_leaves_to_values(
                sub_step.get_global_environment_defaults('env1')
            ),
            {
                'environment-name': 'env1',
                'test1': 'global-env1-default-1',
                'test2': 'global-env1-default-2'
            }
        )
        self.assertEqual(
            ConfigValue.convert_leaves_to_values(
                sub_step.get_global_environment_defaults('env2')
            ),
            {
                'environment-name': 'env2',
                'test1': 'global-env2-default-1',
                'test2': 'global-env2-default-2'
            }
        )

    def test_get_sub_step_env_config(self):
        config = Config({
            Config.CONFIG_KEY: {
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'test1': 'foo'
                        },
                        'environment-config': {
                            'env1': {
                                'test2': 'bar'
                            }
                        }
                    }
                ]

            }
        })

        step_config = config.get_step_config('step-foo')
        sub_step = step_config.get_sub_step('foo1')

        self.assertEqual(
            ConfigValue.convert_leaves_to_values(
                sub_step.get_sub_step_env_config('env1')
            ),
            {
                'test2': 'bar'
            }
        )

    def test_merge_sub_step_config_duplicate_sub_step_keys(self):
        with self.assertRaisesRegex(
                ValueError,
                r"Error merging new sub step configuration into existing sub step configuration for sub step \(foo1\) of step \(step-foo\): Conflict at test1"
        ):
            Config([
                {
                    Config.CONFIG_KEY: {
                        'step-foo': [
                            {
                                'implementer': 'foo1',
                                'config': {
                                    'test1': 'foo'
                                }
                            },
                        ]

                    }
                },
                {
                    Config.CONFIG_KEY: {
                        'step-foo': [
                            {
                                'implementer': 'foo1',
                                'config': {
                                    'test1': 'bar'
                                }
                            },
                        ]

                    }
                }
            ])

    def test_merge_sub_step_config_duplicate_sub_step_environment_config_keys(self):
        with self.assertRaisesRegex(
                ValueError,
                r"Error merging new sub step environment configuration into existing sub step environment configuration for sub step \(foo1\) of step \(step-foo\): Conflict at env1.dup-key"
        ):
            Config([
                {
                    Config.CONFIG_KEY: {
                        'step-foo': [
                            {
                                'implementer': 'foo1',
                                'config': {
                                    'test1': 'foo'
                                },
                                'environment-config': {
                                    'env1': {
                                        'dup-key': 'value1'
                                    }
                                }
                            }
                        ]

                    }
                },
                {
                    Config.CONFIG_KEY: {
                        'step-foo': [
                            {
                                'implementer': 'foo1',
                                'config': {
                                    'test2': 'bar'
                                },
                                'environment-config': {
                                    'env1': {
                                        'dup-key': 'value2'
                                    }
                                }
                            }
                        ]
                    }
                }
            ])

    def test_get_copy_of_runtime_step_config_global_defaults_global_env_defaults_sub_step_config_sub_step_env_config_step_config_overrides(self):
        config = Config({
            Config.CONFIG_KEY: {
                'global-defaults': {
                    'global-default-unique-0': 'global-default',
                    'global-default-override-by-global-env-default-0': 'global-default-override-me',
                    'global-default-override-by-step-config-0': 'global-default-override-me',
                    'global-default-override-by-step-env-config-0': 'global-default-override-me',
                    'global-default-override-by-step-config-override-0': 'global-default-override-me'
                },
                'global-environment-defaults' : {
                    'env1': {
                        'global-default-override-by-global-env-default-0': 'global-environment-defaults-1',
                        'global-env-default-env1-unique-0': 'global-environment-defaults-1',
                        'global-env-default-override-by-step-config': 'global-environment-defaults-1'
                    },
                    'env2': {
                        'global-default-override-by-global-env-default-0': 'global-environment-defaults-2',
                        'global-env-default-env2-unique-0': 'global-environment-defaults-2',
                        'global-env-default-override-by-step-config': 'global-environment-defaults-1'
                    }
                },
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'step-foo-foo1-unique-0': 'step-foo-foo1',
                            'step-foo-foo1-override-by-step-env-config': 'step-foo-foo-override-me',
                            'step-foo-foo1-override-by-step-override': 'step-foo-foo-override-me',
                            'global-default-override-by-step-config-0': 'step-foo-foo1',
                            'global-env-default-override-by-step-config': 'step-foo-foo1'
                        },
                        'environment-config': {
                            'env1': {
                                'step-foo-foo1-override-by-step-env-config': 'step-foo-foo-env1',
                                'global-default-override-by-step-env-config-0': 'step-foo-foo-env1',
                                'step-foo-foo1-env1-unique-0': 'step-foo-foo-env1',
                                'step-foo-foo1-env-specific': 'step-foo-foo-env1'
                            },
                            'env2': {
                                'step-foo-foo1-override-by-step-env-config': 'step-foo-foo-env2',
                                'global-default-override-by-step-env-config-0': 'step-foo-foo-env2',
                                'step-foo-foo1-env2-unique-0': 'step-foo-foo-env2',
                                'step-foo-foo1-env-specific': 'step-foo-foo-env2'
                            }
                        }
                    }
                ]

            }
        })

        config.set_step_config_overrides('step-foo', {
            'global-default-override-by-step-config-override-0': 'step-foo-step-config-override',
            'step-foo-foo1-override-by-step-override' : 'step-foo-step-config-override',
            'step-config-override-unique-0': 'step-config-override-unique'
        })

        step_config = config.get_step_config('step-foo')
        sub_step = step_config.get_sub_step('foo1')

        runtime_step_config_no_given_env = sub_step.get_copy_of_runtime_step_config()
        self.assertEqual(
            ConfigValue.convert_leaves_to_values(runtime_step_config_no_given_env),
            {
                'global-default-unique-0': 'global-default',
                'global-default-override-by-global-env-default-0': 'global-default-override-me',
                'global-default-override-by-step-config-0': 'step-foo-foo1',
                'global-default-override-by-step-env-config-0': 'global-default-override-me',
                'step-foo-foo1-unique-0': 'step-foo-foo1',
                'step-foo-foo1-override-by-step-env-config': 'step-foo-foo-override-me',
                'global-default-override-by-step-config-override-0': 'step-foo-step-config-override',
                'step-foo-foo1-override-by-step-override' : 'step-foo-step-config-override',
                'step-config-override-unique-0': 'step-config-override-unique',
                'global-env-default-override-by-step-config': 'step-foo-foo1'
            }
        )

        runtime_step_config_env1 = sub_step.get_copy_of_runtime_step_config('env1')
        self.assertEqual(
            ConfigValue.convert_leaves_to_values(runtime_step_config_env1),
            {
                'environment-name': 'env1',
                'global-default-unique-0': 'global-default',
                'global-default-override-by-global-env-default-0': 'global-environment-defaults-1',
                'global-default-override-by-step-config-0': 'step-foo-foo1',
                'global-default-override-by-step-env-config-0': 'step-foo-foo-env1',
                'step-foo-foo1-unique-0': 'step-foo-foo1',
                'step-foo-foo1-override-by-step-env-config': 'step-foo-foo-env1',
                'global-default-override-by-step-config-override-0': 'step-foo-step-config-override',
                'step-foo-foo1-override-by-step-override' : 'step-foo-step-config-override',
                'step-config-override-unique-0': 'step-config-override-unique',
                'global-env-default-override-by-step-config': 'step-foo-foo1',
                'global-env-default-env1-unique-0': 'global-environment-defaults-1',
                'step-foo-foo1-env1-unique-0': 'step-foo-foo-env1',
                'step-foo-foo1-env-specific': 'step-foo-foo-env1'
            }
        )

        runtime_step_config_env2 = sub_step.get_copy_of_runtime_step_config('env2')
        self.assertEqual(
            ConfigValue.convert_leaves_to_values(runtime_step_config_env2),
            {
                'environment-name': 'env2',
                'global-default-unique-0': 'global-default',
                'global-default-override-by-global-env-default-0': 'global-environment-defaults-2',
                'global-default-override-by-step-config-0': 'step-foo-foo1',
                'global-default-override-by-step-env-config-0': 'step-foo-foo-env2',
                'step-foo-foo1-unique-0': 'step-foo-foo1',
                'step-foo-foo1-override-by-step-env-config': 'step-foo-foo-env2',
                'global-default-override-by-step-config-override-0': 'step-foo-step-config-override',
                'step-foo-foo1-override-by-step-override' : 'step-foo-step-config-override',
                'step-config-override-unique-0': 'step-config-override-unique',
                'global-env-default-override-by-step-config': 'step-foo-foo1',
                'global-env-default-env2-unique-0': 'global-environment-defaults-2',
                'step-foo-foo1-env2-unique-0': 'step-foo-foo-env2',
                'step-foo-foo1-env-specific': 'step-foo-foo-env2'
            }
        )

    def test_get_copy_of_runtime_step_config_default_global_defaults_global_env_defaults_sub_step_config_sub_step_env_config_step_config_overrides(self):
        config = Config({
            Config.CONFIG_KEY: {
                'global-defaults': {
                    'global-default-unique-0': 'global-default',
                    'global-default-override-by-global-env-default-0': 'global-default-override-me',
                    'global-default-override-by-step-config-0': 'global-default-override-me',
                    'global-default-override-by-step-env-config-0': 'global-default-override-me',
                    'global-default-override-by-step-config-override-0': 'global-default-override-me',
                    'default-overriden-by-global-default': 'global-default',
                },
                'global-environment-defaults' : {
                    'env1': {
                        'global-default-override-by-global-env-default-0': 'global-environment-defaults-1',
                        'global-env-default-env1-unique-0': 'global-environment-defaults-1',
                        'global-env-default-override-by-step-config': 'global-environment-defaults-1',
                        'default-overriden-by-global-env-default': 'global-environment-defaults-1'
                    },
                    'env2': {
                        'global-default-override-by-global-env-default-0': 'global-environment-defaults-2',
                        'global-env-default-env2-unique-0': 'global-environment-defaults-2',
                        'global-env-default-override-by-step-config': 'global-environment-defaults-2',
                        'default-overriden-by-global-env-default': 'global-environment-defaults-2'
                    }
                },
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'step-foo-foo1-unique-0': 'step-foo-foo1',
                            'step-foo-foo1-override-by-step-env-config': 'step-foo-foo-override-me',
                            'step-foo-foo1-override-by-step-override': 'step-foo-foo-override-me',
                            'global-default-override-by-step-config-0': 'step-foo-foo1',
                            'global-env-default-override-by-step-config': 'step-foo-foo1',
                            'default-overriden-by-step-config': 'step-foo-foo1'
                        },
                        'environment-config': {
                            'env1': {
                                'step-foo-foo1-override-by-step-env-config': 'step-foo-foo-env1',
                                'global-default-override-by-step-env-config-0': 'step-foo-foo-env1',
                                'step-foo-foo1-env1-unique-0': 'step-foo-foo-env1',
                                'step-foo-foo1-env-specific': 'step-foo-foo-env1',
                                'default-overriden-by-step-env-config': 'step-foo-foo-env1',
                            },
                            'env2': {
                                'step-foo-foo1-override-by-step-env-config': 'step-foo-foo-env2',
                                'global-default-override-by-step-env-config-0': 'step-foo-foo-env2',
                                'step-foo-foo1-env2-unique-0': 'step-foo-foo-env2',
                                'step-foo-foo1-env-specific': 'step-foo-foo-env2',
                                'default-overriden-by-step-env-config': 'step-foo-foo-env2',
                            }
                        }
                    }
                ]

            }
        })

        config.set_step_config_overrides('step-foo', {
            'global-default-override-by-step-config-override-0': 'step-foo-step-config-override',
            'step-foo-foo1-override-by-step-override' : 'step-foo-step-config-override',
            'step-config-override-unique-0': 'step-config-override-unique',
            'default-overriden-by-step-config-overrides': 'step-foo-step-config-override'
        })

        step_config = config.get_step_config('step-foo')
        sub_step = step_config.get_sub_step('foo1')

        defaults = {
            'default-unique': 'not-overriden-default',
            'default-overriden-by-global-default': 'override-me',
            'default-overriden-by-global-env-default': 'override-me',
            'default-overriden-by-step-config': 'override-me',
            'default-overriden-by-step-env-config': 'override-me',
            'default-overriden-by-step-config-overrides': 'override-me'
        }

        runtime_step_config_no_given_env = sub_step.get_copy_of_runtime_step_config(None, defaults)
        self.assertEqual(
            ConfigValue.convert_leaves_to_values(runtime_step_config_no_given_env),
            {
                'global-default-unique-0': 'global-default',
                'global-default-override-by-global-env-default-0': 'global-default-override-me',
                'global-default-override-by-step-config-0': 'step-foo-foo1',
                'global-default-override-by-step-env-config-0': 'global-default-override-me',
                'step-foo-foo1-unique-0': 'step-foo-foo1',
                'step-foo-foo1-override-by-step-env-config': 'step-foo-foo-override-me',
                'global-default-override-by-step-config-override-0': 'step-foo-step-config-override',
                'step-foo-foo1-override-by-step-override' : 'step-foo-step-config-override',
                'step-config-override-unique-0': 'step-config-override-unique',
                'global-env-default-override-by-step-config': 'step-foo-foo1',
                'default-unique': 'not-overriden-default',
                'default-overriden-by-global-default': 'global-default',
                'default-overriden-by-global-env-default': 'override-me',
                'default-overriden-by-step-config': 'step-foo-foo1',
                'default-overriden-by-step-env-config': 'override-me',
                'default-overriden-by-step-config-overrides': 'step-foo-step-config-override'
            }
        )

        runtime_step_config_env1 = sub_step.get_copy_of_runtime_step_config('env1', defaults)
        self.assertEqual(
            ConfigValue.convert_leaves_to_values(runtime_step_config_env1),
            {
                'environment-name': 'env1',
                'global-default-unique-0': 'global-default',
                'global-default-override-by-global-env-default-0': 'global-environment-defaults-1',
                'global-default-override-by-step-config-0': 'step-foo-foo1',
                'global-default-override-by-step-env-config-0': 'step-foo-foo-env1',
                'step-foo-foo1-unique-0': 'step-foo-foo1',
                'step-foo-foo1-override-by-step-env-config': 'step-foo-foo-env1',
                'global-default-override-by-step-config-override-0': 'step-foo-step-config-override',
                'step-foo-foo1-override-by-step-override' : 'step-foo-step-config-override',
                'step-config-override-unique-0': 'step-config-override-unique',
                'global-env-default-override-by-step-config': 'step-foo-foo1',
                'global-env-default-env1-unique-0': 'global-environment-defaults-1',
                'step-foo-foo1-env1-unique-0': 'step-foo-foo-env1',
                'step-foo-foo1-env-specific': 'step-foo-foo-env1',
                'default-unique': 'not-overriden-default',
                'default-overriden-by-global-default': 'global-default',
                'default-overriden-by-global-env-default': 'global-environment-defaults-1',
                'default-overriden-by-step-config': 'step-foo-foo1',
                'default-overriden-by-step-env-config': 'step-foo-foo-env1',
                'default-overriden-by-step-config-overrides': 'step-foo-step-config-override'
            }
        )

        runtime_step_config_env2 = sub_step.get_copy_of_runtime_step_config('env2', defaults)
        self.assertEqual(
            ConfigValue.convert_leaves_to_values(runtime_step_config_env2),
            {
                'environment-name': 'env2',
                'global-default-unique-0': 'global-default',
                'global-default-override-by-global-env-default-0': 'global-environment-defaults-2',
                'global-default-override-by-step-config-0': 'step-foo-foo1',
                'global-default-override-by-step-env-config-0': 'step-foo-foo-env2',
                'step-foo-foo1-unique-0': 'step-foo-foo1',
                'step-foo-foo1-override-by-step-env-config': 'step-foo-foo-env2',
                'global-default-override-by-step-config-override-0': 'step-foo-step-config-override',
                'step-foo-foo1-override-by-step-override' : 'step-foo-step-config-override',
                'step-config-override-unique-0': 'step-config-override-unique',
                'global-env-default-override-by-step-config': 'step-foo-foo1',
                'global-env-default-env2-unique-0': 'global-environment-defaults-2',
                'step-foo-foo1-env2-unique-0': 'step-foo-foo-env2',
                'step-foo-foo1-env-specific': 'step-foo-foo-env2',
                'default-unique': 'not-overriden-default',
                'default-overriden-by-global-default': 'global-default',
                'default-overriden-by-global-env-default': 'global-environment-defaults-2',
                'default-overriden-by-step-config': 'step-foo-foo1',
                'default-overriden-by-step-env-config': 'step-foo-foo-env2',
                'default-overriden-by-step-config-overrides': 'step-foo-step-config-override'
            }
        )

    def test_get_config_value(self):
        config = Config({
            Config.CONFIG_KEY: {
                'global-defaults': {
                    'global-default-unique-0': 'global-default',
                    'global-default-override-by-global-env-default-0': 'global-default-override-me',
                    'global-default-override-by-step-config-0': 'global-default-override-me',
                    'global-default-override-by-step-env-config-0': 'global-default-override-me',
                    'global-default-override-by-step-config-override-0': 'global-default-override-me'
                },
                'global-environment-defaults' : {
                    'env1': {
                        'global-default-override-by-global-env-default-0': 'global-environment-defaults-1',
                        'global-env-default-env1-unique-0': 'global-environment-defaults-1',
                        'global-env-default-override-by-step-config': 'global-environment-defaults-1'
                    },
                    'env2': {
                        'global-default-override-by-global-env-default-0': 'global-environment-defaults-2',
                        'global-env-default-env2-unique-0': 'global-environment-defaults-2',
                        'global-env-default-override-by-step-config': 'global-environment-defaults-1'
                    }
                },
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'step-foo-foo1-unique-0': 'step-foo-foo1',
                            'step-foo-foo1-override-by-step-env-config': 'step-foo-foo-override-me',
                            'step-foo-foo1-override-by-step-override': 'step-foo-foo-override-me',
                            'global-default-override-by-step-config-0': 'step-foo-foo1',
                            'global-env-default-override-by-step-config': 'step-foo-foo1'
                        },
                        'environment-config': {
                            'env1': {
                                'step-foo-foo1-override-by-step-env-config': 'step-foo-foo-env1',
                                'global-default-override-by-step-env-config-0': 'step-foo-foo-env1',
                                'step-foo-foo1-env1-unique-0': 'step-foo-foo-env1',
                                'step-foo-foo1-env-specific': 'step-foo-foo-env1'
                            },
                            'env2': {
                                'step-foo-foo1-override-by-step-env-config': 'step-foo-foo-env2',
                                'global-default-override-by-step-env-config-0': 'step-foo-foo-env2',
                                'step-foo-foo1-env2-unique-0': 'step-foo-foo-env2',
                                'step-foo-foo1-env-specific': 'step-foo-foo-env2'
                            }
                        }
                    }
                ]

            }
        })

        config.set_step_config_overrides('step-foo', {
            'global-default-override-by-step-config-override-0': 'step-foo-step-config-override',
            'step-foo-foo1-override-by-step-override' : 'step-foo-step-config-override',
            'step-config-override-unique-0': 'step-config-override-unique'
        })

        step_config = config.get_step_config('step-foo')
        sub_step = step_config.get_sub_step('foo1')

        self.assertEqual(
            sub_step.get_config_value('global-default-unique-0'),
            "global-default")

        self.assertEqual(
            sub_step.get_config_value('global-default-override-by-global-env-default-0'),
            "global-default-override-me")

        self.assertEqual(
            sub_step.get_config_value('global-default-override-by-global-env-default-0', 'env1'),
            "global-environment-defaults-1")
        self.assertEqual(
            sub_step.get_config_value('global-default-override-by-global-env-default-0', 'env2'),
            "global-environment-defaults-2")

        self.assertEqual(
            sub_step.get_config_value('global-default-override-by-step-config-0'),
            "step-foo-foo1")
        self.assertEqual(
            sub_step.get_config_value('global-default-override-by-step-config-0', 'env1'),
            "step-foo-foo1")

        self.assertEqual(
            sub_step.get_config_value('step-foo-foo1-unique-0',),
            "step-foo-foo1")
        self.assertEqual(
            sub_step.get_config_value('step-foo-foo1-unique-0', 'env1'),
            "step-foo-foo1")

        self.assertEqual(
            sub_step.get_config_value('step-foo-foo1-override-by-step-env-config',),
            "step-foo-foo-override-me")
        self.assertEqual(
            sub_step.get_config_value('step-foo-foo1-override-by-step-env-config', 'env1'),
            "step-foo-foo-env1")
        self.assertEqual(
            sub_step.get_config_value('step-foo-foo1-override-by-step-env-config', 'env2'),
            "step-foo-foo-env2")

        self.assertIsNone(sub_step.get_config_value('does-not-exist'))

    def test_get_config_value_list(self):
        config = Config({
            Config.CONFIG_KEY: {
                'global-defaults': {
                    'global-default-unique-0': ['global-default-a', 'global-default-b'],
                },
                'step-foo': [
                    {
                        'implementer': 'foo1',
                        'config': {
                            'step-foo-foo1-unique-0': ['step-foo-foo1-a', 'step-foo-foo1-b'],
                        }
                    }
                ]

            }
        })

        step_config = config.get_step_config('step-foo')
        sub_step = step_config.get_sub_step('foo1')

        self.assertEqual(
            sub_step.get_config_value('global-default-unique-0'),
            ['global-default-a', 'global-default-b']
        )

        self.assertEqual(
            sub_step.get_config_value('step-foo-foo1-unique-0'),
            ['step-foo-foo1-a', 'step-foo-foo1-b']
        )



# pylint: disable=too-many-lines
"""`StepImplementer` for the `deploy` step using ArgoCD.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:

  * static configuration
  * runtime configuration
  * previous step results

Configuration Key         | Required? | Default  | Description
--------------------------|-----------|----------|---------------------------
`argocd-username`         | Yes       |          | Username for accessing the ArgoCD API
`argocd-password`         | Yes       |          | Password for accessing the ArgoCD API
`argocd-api`              | Yes       |          | The ArgoCD API endpoint
`argocd-auto-sync`        | Yes       | True     | If set to false, argo cd will sync only if \
                                                   explicitly told to do so via the UI or CLI. \
                                                   Otherwise it will sync if the repo contents \
                                                   have changed
`argocd-skip-tls`         | Yes       | False    | `False` to not ignore TLS issues when \
                                                   authenticating with ArgoCD. True` to ignore TLS \
                                                   issues when authenticating with ArgoCD.
`argocd-sync-timeout-seconds` \
                          | Yes       | 60       | Number of seconds to wait for argocd to \
                                                   sync updates
`argocd-sync-retry-limit` | No        | 3        | Value to pass to retry limit flag for argo sync
`deployment-config-repo`  | Yes       |          | The repo containing the helm chart definition
`deployment-config-helm-chart-path` \
                          | Yes       | ./       | Directory containing the helm chart definition
`deployment-config-helm-chart-environment-values-file` \
                          | No        | values-{ENV}.yaml | File to update with environment \
                                                            deployment specific information. \
                                                            <br/>\
                                                            Default value uses `environment` \
                                                            configuration option to calculate file \
                                                            name.
`deployment-config-helm-chart-additional-values-files` \
                          | No        |          | Array of additional values files. \
                                                   Paths must be relative to the given \
                                                   Helm Chart path (`deployment-config-helm-chart-path`). \
                                                   <br/>\
                                                   Does not need to include \
                                                   `deployment-config-helm-chart-environment-values-file` \
                                                   as it will be automatically included as the \
                                                   last `--values` flag.
                                                   <br/>\
                                                   Used when createing the ArgoCD Application.\
                                                   <br/><br/>\
                                                   **NOTE:** Helm will always pick up `values.yaml`\
                                                   by default whether or not additional values \
                                                   files are provided, therefor no need to \
                                                   explicitly provide.
`deployment-config-helm-chart-values-file-image-tag-yq-path` \
                          | Yes       | `image_tag` | YQ path to value in `deployment-config-helm-chart-environment-values-file` \
                                                      to update with the `container-image-tag` \
                                                      before deployment. \
                                                      <br/>\
                                                      **SEE:**: https://github.com/mikefarah/yq \
                                                      for documentation on valid yq paths.
`kube-api-uri`            | Yes       | https://kubernetes.default.svc | k8s API endpoint
`kube-api-token`          | No        |          | k8s API token. This is used to add an external \
                                                   k8s cluster into argocd. It is required if the \
                                                   cluster has not already been added to ArgoCD. \
                                                   The token should be persistent \
                                                   (.e.g, a service account token) and have cluster \
                                                   admin access.
`kube-api-skip-tls`       | Yes       | False    | Whether or not to skip tls verification when \
                                                   authenticating to an external k8s cluster. \
                                                   Used when a new cluster is registered with \
                                                   argocd.
`git-email`               | Yes       |          | Git email for commit
`git-name`                | Yes       | `Ploigos Robot` | Git name for commit
`git-username`            | No        |          | If the helm config repo s accessed via http(s) \
                                                   this must be supplied
`git-password`            | No        |          | If the helm config repo is accessed via http(s) \
                                                   this must be supplied
`tag`                     | No        | latest   | The git tag to apply to the config repo. \
                                                   If not supplied `version` will be used. \
                                                   If `version` not supplied `latest` will be used.
`version`                 | No        | latest   | Ignored if `tag` is provided. \
                                                   The git tag to apply to the config repo if `tag` \
                                                   is not supplied. \
                                                   If `tag` and `version` not supplied `latest` \
                                                   will be used.
`container-image-tag`     | Yes       |          | Tag container image was pushed with. <br/>\
                                                   Takes the form of: \
                                                     "`container-image-registry-uri`\
                                                        /`container-image-registry-organization`\
                                                        /`container-image-repository`\
                                                        :`container-image-version`"
`force-push-tags`         | No        | False    | Force push Git Tags
`additional-helm-values-files` \
                          | No    | []       | Array of value files to add to argocd app for helm use

Results
-------
Results output by this step.

Result Key                 | Description
---------------------------|------------
`argocd-app-name`          | The argocd app name that was created or updated
`deployed-host-urls`       | The host URLs deployed by ArgoCD (Ingress/Route resources)
`config-repo-git-tag`      | The git tag applied to the configuration repo for deployment
`argocd-deployed-manifest` | The generated yml file used for deployment.
""" # pylint: disable=line-too-long
import os
import re
import sys

import sh
import yaml
from ploigos_step_runner import StepImplementer
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner import StepResult

DEFAULT_CONFIG = {
    'argocd-sync-timeout-seconds': 60,
    'argocd-sync-retry-limit': 3,
    'argocd-auto-sync': True,
    'argocd-skip-tls' : False,
    'deployment-config-helm-chart-path': './',
    'deployment-config-helm-chart-additional-values-files': [],
    'additional-helm-values-files': [],
    'deployment-config-helm-chart-values-file-image-tag-yq-path': 'image_tag',
    'force-push-tags': False,
    'kube-api-skip-tls': False,
    'kube-api-uri': 'https://kubernetes.default.svc',
    'git-name': 'Ploigos Robot'
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'argocd-username',
    'argocd-password',
    'argocd-api',
    'argocd-skip-tls',
    'deployment-config-repo',
    'deployment-config-helm-chart-path',
    'deployment-config-helm-chart-values-file-image-tag-yq-path',
    'git-email',
    'git-name',
    'container-image-tag'
]

GIT_AUTHENTICATION_CONFIG = {
    'git-username': None,
    'git-password': None
}

KUBE_LABEL_NOT_SAFE_CHARS_REGEX = r"[^a-zA-Z0-9\-_\.]"
KUBE_LABEL_NOT_SAFE_BEGINING_END_CHARS_REGEX = r"^[^a-zA-Z0-9]*|[^a-zA-Z0-9]*$"
KUBE_LABEL_MAX_LENGTH = 52
KUBE_LABEL_REPLACEMENT_CHAR = '-'


class ArgoCD(StepImplementer):
    """`StepImplementer` for the `deploy` step using ArgoCD.
    """

    GIT_REPO_REGEX = re.compile(r"(?P<protocol>^https:\/\/|^http:\/\/)?(?P<address>.*$)")

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.
        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _validate_required_config_or_previous_step_result_artifact_keys(self):
        """Validates that the required configuration keys or previous step result artifacts
        are set and have valid values.

        Validates that:
        * required configuration is given
        * either both git-username and git-password are set or neither.

        Raises
        ------
        StepRunnerException
            If step configuration or previous step result artifacts have invalid required values
        """
        super()._validate_required_config_or_previous_step_result_artifact_keys()

        # ensure either both git-username and git-password are set or neither
        runtime_auth_config = {}
        for auth_config_key in GIT_AUTHENTICATION_CONFIG:
            runtime_auth_config_value = self.get_value(auth_config_key)

            if runtime_auth_config_value is not None:
                runtime_auth_config[auth_config_key] = runtime_auth_config_value

        if (any(element in runtime_auth_config for element in GIT_AUTHENTICATION_CONFIG)) and \
                (not all(element in runtime_auth_config for element in GIT_AUTHENTICATION_CONFIG)):
            raise StepRunnerException(
                "Either 'git-username' or 'git-password 'is not set. Neither or both must be set."
            )

        # ensure if deployment config repo is using http/https that user/pass is provided
        deployment_config_repo_uri = self.get_value('deployment-config-repo')
        if re.match(r'^http://|^https://', deployment_config_repo_uri) and not runtime_auth_config:
            raise StepRunnerException(
                f"Since provided 'deployment-config-repo' ({deployment_config_repo_uri}) uses"
                f" http/https protical both 'git-username' and 'git-password' must be provided."
            )

    def _run_step(self):  # pylint: disable=too-many-locals
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        # get input
        deployment_config_repo = self.get_value('deployment-config-repo')
        deployment_config_repo_branch = ArgoCD.__get_repo_branch()
        deployment_config_helm_chart_path = self.get_value('deployment-config-helm-chart-path')
        deployment_config_destination_cluster_uri = self.get_value('kube-api-uri')
        deployment_config_destination_cluster_token = self.get_value('kube-api-token')
        deployment_config_helm_chart_environment_values_file = \
            self.__get_deployment_config_helm_chart_environment_values_file()
        deployment_config_helm_chart_values_file_image_tag_yq_path = \
            self.get_value('deployment-config-helm-chart-values-file-image-tag-yq-path')
        deployment_config_helm_chart_additional_value_files = \
            self.get_value('deployment-config-helm-chart-additional-values-files')
        container_image_tag = self.get_value('container-image-tag')
        force_push_tags = self.get_value('force-push-tags')
        additional_helm_values_files = self.get_value('additional-helm-values-files')

        try:
            argocd_app_name = self.__get_app_name()
            step_result.add_artifact(
                name='argocd-app-name',
                value=argocd_app_name
            )

            # clone the configuration repository
            print("Clone the configuration repository")
            clone_repo_dir_name = 'deployment-config-repo'
            deployment_config_repo_dir = ArgoCD.__clone_repo(
                repo_dir=self.create_working_dir_sub_dir(clone_repo_dir_name),
                repo_url=deployment_config_repo,
                repo_branch=deployment_config_repo_branch,
                git_email=self.get_value('git-email'),
                git_name=self.get_value('git-name'),
                username = self.get_value('git-username'),
                password = self.get_value('git-password')
            )

            # update values file, commit it, push it, and tag it
            print("Update the environment values file")
            deployment_config_helm_chart_environment_values_file_path = os.path.join(
                deployment_config_repo_dir,
                deployment_config_helm_chart_path,
                deployment_config_helm_chart_environment_values_file
            )
            self.__update_yaml_file_value(
                file=deployment_config_helm_chart_environment_values_file_path,
                yq_path=deployment_config_helm_chart_values_file_image_tag_yq_path,
                value=container_image_tag
            )
            print("Commit the updated environment values file")
            ArgoCD.__git_commit_file(
                git_commit_message=f'Updating values for deployment to {self.environment}',
                file_path=os.path.join(
                    deployment_config_helm_chart_path,
                    deployment_config_helm_chart_environment_values_file
                ),
                repo_dir=deployment_config_repo_dir
            )
            print("Tag and push the updated environment values file")
            deployment_config_repo_tag = self.__get_deployment_config_repo_tag()
            self.__git_tag_and_push_deployment_config_repo(
                deployment_config_repo=deployment_config_repo,
                deployment_config_repo_dir=deployment_config_repo_dir,
                deployment_config_repo_tag=deployment_config_repo_tag,
                force_push_tags=force_push_tags
            )
            step_result.add_artifact(
                name='config-repo-git-tag',
                value=deployment_config_repo_tag
            )

            # create/update argocd app and sync it
            print("Sign into ArgoCD")
            ArgoCD.__argocd_sign_in(
                argocd_api=self.get_value('argocd-api'),
                username=self.get_value('argocd-username'),
                password=self.get_value('argocd-password'),
                insecure=self.get_value('argocd-skip-tls')
            )
            print("Add target cluster to ArgoCD")
            self.__argocd_add_target_cluster(
                kube_api=deployment_config_destination_cluster_uri,
                kube_api_token=deployment_config_destination_cluster_token,
                kube_api_skip_tls=self.get_value('kube-api-skip-tls')
            )
            print(f"Create or update ArgoCD Application ({argocd_app_name})")
            argocd_values_files = []
            argocd_values_files += deployment_config_helm_chart_additional_value_files
            argocd_values_files += [deployment_config_helm_chart_environment_values_file]
            argocd_values_files += additional_helm_values_files
            ArgoCD.__argocd_app_create_or_update(
                argocd_app_name=argocd_app_name,
                repo=deployment_config_repo,
                revision=deployment_config_repo_branch,
                path=deployment_config_helm_chart_path,
                dest_server=deployment_config_destination_cluster_uri,
                auto_sync=self.get_value('argocd-auto-sync'),
                values_files=argocd_values_files
            )

            # sync and wait for the sync of the ArgoCD app
            print(f"Sync (and wait for) ArgoCD Application ({argocd_app_name})")
            ArgoCD.__argocd_app_sync(
                argocd_app_name=argocd_app_name,
                argocd_sync_timeout_seconds=self.get_value('argocd-sync-timeout-seconds'),
                argocd_sync_retry_limit=self.get_value('argocd-sync-retry-limit')
            )

            # get the ArgoCD app manifest that was synced
            print(f"Get ArgoCD Application ({argocd_app_name}) synced manifest")
            argocd_app_manifest_file = self.__argocd_get_app_manifest(
                argocd_app_name=argocd_app_name
            )
            step_result.add_artifact(
                name='argocd-deployed-manifest',
                value=argocd_app_manifest_file
            )

            # determine the deployed host URLs
            print(
                "Determine the deployed host URLs for the synced"
                f" ArgoCD Application (({argocd_app_name})"
            )
            deployed_host_urls = ArgoCD.__get_deployed_host_urls(
                manifest_path=argocd_app_manifest_file
            )
            step_result.add_artifact(
                name='deployed-host-urls',
                value=deployed_host_urls
            )
        except StepRunnerException as error:
            step_result.success = False
            step_result.message = f"Error deploying to environment ({self.environment}):" \
                f" {str(error)}"

        return step_result

    def __get_deployment_config_helm_chart_environment_values_file(self):
        """Get the deployment configuration Helm Chart environment specific value file.

        Returns
        -------
        str
            Deployment configuration Helm Chart environment specific value file.
        """
        deployment_config_helm_chart_env_value_file = \
            self.get_value('deployment-config-helm-chart-environment-values-file')

        if not deployment_config_helm_chart_env_value_file:
            if self.environment:
                deployment_config_helm_chart_env_value_file = f'values-{self.environment}.yaml'
            else:
                deployment_config_helm_chart_env_value_file = 'values.yaml'

        return deployment_config_helm_chart_env_value_file

    def __update_yaml_file_value(self, file, yq_path, value):
        """Update a YAML file using YQ.

        Parameters
        ----------
        file : str
            Path to file to update.
        yq_path : str
            YQ path to the value to update.
        value: str
            value to update the `yq_path`

        Returns
        -------
        str
            Path to the file to update.

        Raises
        ------
        StepRunnerException
            If error updating file.
        """
        # NOTE: use a YQ script to update so that comment can be injected
        yq_script_file = self.write_working_file(
            filename='update-yaml-file-yq-script.yaml',
            contents=bytes(
                f"- command: update\n"
                f"  path: {yq_path}\n"
                f"  value:\n"
                f"    {value} # written by ploigos-step-runner\n",
                'utf-8'
            )
        )

        # inplace update the file
        try:
            sh.yq.write( # pylint: disable=no-member
                file,
                f'--script={yq_script_file}',
                '--inplace'
            )
        except sh.ErrorReturnCode as error:
            raise StepRunnerException(
                f"Error updating YAML file ({file}) target ({yq_path}) with value ({value}):"
                f" {error}"
            ) from error

        return file

    def __git_tag_and_push_deployment_config_repo(
        self,
        deployment_config_repo,
        deployment_config_repo_dir,
        deployment_config_repo_tag,
        force_push_tags
    ):
        deployment_config_repo_match = ArgoCD.GIT_REPO_REGEX.match(deployment_config_repo)
        deployment_config_repo_protocol = deployment_config_repo_match.groupdict()['protocol']
        deployment_config_repo_address = deployment_config_repo_match.groupdict()['address']

        # if deployment config repo uses http/https push using user/pass
        # else push using ssh
        if deployment_config_repo_protocol and re.match(
            r'^http://|^https://',
            deployment_config_repo_protocol
        ):
            username = self.get_value('git-username')
            password = self.get_value('git-password')

            deployment_config_repo_with_user_pass = \
                f"{deployment_config_repo_protocol}{username}:{password}" \
                f"@{deployment_config_repo_address}"
            ArgoCD.__git_tag_and_push(
                repo_dir=deployment_config_repo_dir,
                tag=deployment_config_repo_tag,
                url=deployment_config_repo_with_user_pass,
                force_push_tags=force_push_tags
            )
        else:
            ArgoCD.__git_tag_and_push(
                repo_dir=deployment_config_repo_dir,
                tag=deployment_config_repo_tag,
                force_push_tags=force_push_tags
            )

    def __get_app_name(self):
        repo_branch = ArgoCD.__get_repo_branch()
        organization = self.get_value('organization')
        application = self.get_value('application-name')
        service = self.get_value('service-name')
        app_name = f"{organization}-{application}-{service}-{repo_branch}"

        if self.environment:
            app_name = app_name + '-' + self.environment

        # replace dangerous characters in app name
        app_name = app_name.lower()
        app_name = re.sub(
            KUBE_LABEL_NOT_SAFE_CHARS_REGEX,
            KUBE_LABEL_REPLACEMENT_CHAR,
            app_name
        )

        # max length for a kube label / resource name is 63
        if len(app_name) > KUBE_LABEL_MAX_LENGTH:
            app_name = app_name[len(app_name) - KUBE_LABEL_MAX_LENGTH:]

        # be sure app name doesn't start or end with not safe chars
        app_name = re.sub(
            KUBE_LABEL_NOT_SAFE_BEGINING_END_CHARS_REGEX,
            '',
            app_name
        )

        return app_name

    def __get_deployment_config_repo_tag(self):
        """
        Returns
        -------
        str
            String to tag the deployment configuration repository with.

        Raise
        -----
        StepRunnerException
            If i
        """
        tag = self.get_value('tag')

        if tag is None:
            tag = self.get_value('version')

        if tag is None:
            tag = 'latest'
            print(
                "No specified 'tag' or 'version' using 'latest' for"
                "deployment configuration repository tag."
            )

        # need the config repo tag to be unique per environment to prevent conflicts
        if self.environment:
            tag += f".{self.environment}"

        return tag

    @staticmethod
    def __get_deployed_host_urls( # pylint: disable=too-many-branches,too-many-nested-blocks
        manifest_path
    ):
        """Gets the ingress hosts URLs from a manifest of Kubernetes resources.

        Supports:

        - route.openshift.io/v1/Route
        - networking.k8s.io/v1/Ingress

        Returns
        -------
        list of str
            Ingress hosts URLs defined in the given manifest of Kubernetes resources.

        See
        ---
        * https://docs.openshift.com/container-platform/4.6/rest_api/network_apis/ingress-networking-k8s-io-v1.html
        * https://docs.openshift.com/container-platform/4.6/rest_api/network_apis/route-route-openshift-io-v1.html
        """ # pylint: disable=line-too-long
        host_urls = []
        manifest_resources = {}
        # load the manifest
        with open(manifest_path, encoding='utf-8') as file:
            manifest_resources = yaml.load_all(file, Loader=yaml.FullLoader)

            # for each resource in the manfest,
            # determine if its a known type and then attempt to get host and TLS config from it
            for manifest_resource in manifest_resources:
                if manifest_resource is None or 'kind' not in manifest_resource:
                    continue

                kind = manifest_resource['kind']
                api_version = manifest_resource['apiVersion']

                # if Route resource
                if kind == 'Route' and api_version == 'route.openshift.io/v1':
                    # get host
                    if 'host' in manifest_resource['spec']:
                        host = manifest_resource['spec']['host']

                        # determine if TLS route
                        tls = False
                        if 'tls' in manifest_resource['spec']:
                            tls_config = manifest_resource['spec']['tls']
                            if tls_config:
                                tls = True

                        # determine protocol
                        protocol = ''
                        if tls:
                            protocol = 'https://'
                        else:
                            protocol = 'http://'

                        # record the host url
                        host_urls.append(f"{protocol}{host}")

                # if Ingress resource
                if kind == 'Ingress' and api_version == 'networking.k8s.io/v1':
                    ingress_rules = manifest_resource['spec']['rules']
                    for rule in ingress_rules:
                        # get host
                        if 'host' in rule:
                            host = rule['host']

                            # determine if TLS ingress
                            tls = False
                            if 'tls' in manifest_resource['spec']:
                                for tls_config in manifest_resource['spec']['tls']:
                                    if ('hosts' in tls_config) and (host in tls_config['hosts']):
                                        tls = True
                                        break

                            # determine protocol
                            protocol = ''
                            if tls:
                                protocol = 'https://'
                            else:
                                protocol = 'http://'

                            # record the host url
                            host_urls.append(f"{protocol}{host}")

        return host_urls

    @staticmethod
    def __clone_repo( # pylint: disable=too-many-arguments
        repo_dir,
        repo_url,
        repo_branch,
        git_email,
        git_name,
	  username=None,
	  password=None
    ):
        """Clones and checks out the deployment configuration repository.

        Parameters
        ----------
        repo_dir : str
            Path to where to clone the repository
        repo_uri : str
            URI of the repository to clone.
        git_email : str
            email to use when performing git operations in the cloned repository
        git_name : str
            name to use when performing git operations in the cloned repository

        Returns
        -------
        str
            Path to the directory where the deployment configuration repository was cloned
            and checked out.

        Raises
        ------
        StepRunnerException
        * if error cloning repository
        * if error checking out branch of repository
        * if error configuring repo user
        """
        repo_match = ArgoCD.GIT_REPO_REGEX.match(repo_url)
        repo_protocol = repo_match.groupdict()['protocol']
        repo_address = repo_match.groupdict()['address']
        # if deployment config repo uses http/https push using user/pass
        # else push using ssh
        if username and password and repo_protocol and re.match(
            r'^http://|^https://',
            repo_protocol
        ):
            repo_url_with_auth = \
                f"{repo_protocol}{username}:{password}" \
                f"@{repo_address}"
        else:
            repo_url_with_auth = repo_url
        try:
            sh.git.clone( # pylint: disable=no-member
                repo_url_with_auth,
                repo_dir,
                _out=sys.stdout,
                _err=sys.stderr
            )
        except sh.ErrorReturnCode as error:
            raise StepRunnerException(
                f"Error cloning repository ({repo_url}): {error}"
            ) from error

        try:
            # no atomic way in git to checkout out new or existing branch,
            # so first try to check out existing, if that doesn't work try new
            try:
                sh.git.checkout(  # pylint: disable=no-member
                    repo_branch,
                    _cwd=repo_dir,
                    _out=sys.stdout,
                    _err=sys.stderr
                )
            except sh.ErrorReturnCode:
                sh.git.checkout(
                    '-b',
                    repo_branch,
                    _cwd=repo_dir,
                    _out=sys.stdout,
                    _err=sys.stderr
                )
        except sh.ErrorReturnCode as error:
            # NOTE: this should never happen
            raise StepRunnerException(
                f"Unexpected error checking out new or existing branch ({repo_branch})"
                f" from repository ({repo_url}): {error}"
            ) from error

        try:
            sh.git.config( # pylint: disable=no-member
                'user.email',
                git_email,
                _cwd=repo_dir,
                _out=sys.stdout,
                _err=sys.stderr
            )
            sh.git.config( # pylint: disable=no-member
                'user.name',
                git_name,
                _cwd=repo_dir,
                _out=sys.stdout,
                _err=sys.stderr
            )
        except sh.ErrorReturnCode as error:
            # NOTE: this should never happen
            raise StepRunnerException(
                f"Unexpected error configuring git user.email ({git_email})"
                f" and user.name ({git_name}) for repository ({repo_url})"
                f" in directory ({repo_dir}): {error}"
            ) from error

        return repo_dir

    @staticmethod
    def __get_repo_branch():
        try:
            return sh.git( # pylint: disable=no-member,too-many-function-args
                'rev-parse',
                '--abbrev-ref',
                'HEAD'
            ).rstrip()
        except sh.ErrorReturnCode as error:
            # NOTE: this should never happen
            raise StepRunnerException(
                "Unexpected error getting checkedout git repository branch"
                f" of the working directory: {error}"
            ) from error

    @staticmethod
    def __git_tag_and_push(repo_dir, tag, url=None, force_push_tags=False):
        """
        Raises
        ------
        StepRunnerException
        * if error pushing commits
        * if error tagging repository
        * if error pushing tags
        """

        git_push = sh.git.push.bake(url) if url else sh.git.push

        # push commits
        try:
            git_push(
                _cwd=repo_dir,
                _out=sys.stdout
            )
        except sh.ErrorReturnCode as error:
            raise StepRunnerException(
                f"Error pushing commits from repository directory ({repo_dir}) to"
                f" repository ({url}): {error}"
            ) from error

        # tag
        try:
            # NOTE:
            # this force is only needed locally in case of a re-run of the same pipeline
            # without a fresh check out. You will notice there is no force on the push
            # making this an acceptable work around to the issue since on the off chance
            # actually overwriting a tag with a different comment, the push will fail
            # because the tag will be attached to a different git hash.
            sh.git.tag(  # pylint: disable=no-member
                tag,
                '-f',
                _cwd=repo_dir,
                _out=sys.stdout,
                _err=sys.stderr
            )
        except sh.ErrorReturnCode as error:
            raise StepRunnerException(
                f"Error tagging repository ({repo_dir}) with tag ({tag}): {error}"
            ) from error

        git_push_additional_arguments = []
        if force_push_tags:
            git_push_additional_arguments += ["--force"]

        # push tag
        try:
            git_push(
                '--tag',
                *git_push_additional_arguments,
                _cwd=repo_dir,
                _out=sys.stdout
            )
        except sh.ErrorReturnCode as error:
            raise StepRunnerException(
                f"Error pushing tags from repository directory ({repo_dir}) to"
                f" repository ({url}): {error}"
            ) from error

    @staticmethod
    def __git_commit_file(git_commit_message, file_path, repo_dir):
        try:
            sh.git.add( # pylint: disable=no-member
                file_path,
                _cwd=repo_dir,
                _out=sys.stdout,
                _err=sys.stderr
            )
        except sh.ErrorReturnCode as error:
            # NOTE: this should never happen
            raise StepRunnerException(
                f"Unexpected error adding file ({file_path}) to commit"
                f" in git repository ({repo_dir}): {error}"
            ) from error

        try:
            sh.git.commit( # pylint: disable=no-member
                '--allow-empty',
                '--all',
                '--message', git_commit_message,
                _cwd=repo_dir,
                _out=sys.stdout,
                _err=sys.stderr
            )
        except sh.ErrorReturnCode as error:
            # NOTE: this should never happen
            raise StepRunnerException(
                f"Unexpected error commiting file ({file_path})"
                f" in git repository ({repo_dir}): {error}"
            ) from error

    @staticmethod
    def __argocd_sign_in(
        argocd_api,
        username,
        password,
        insecure=False
    ):
        """Signs into ArgoCD CLI.

        Raises
        ------
        StepRunnerException
            If error signing into ArgoCD CLI.
        """
        try:
            insecure_flag = None
            if insecure:
                insecure_flag = '--insecure'

            sh.argocd.login(  # pylint: disable=no-member
                argocd_api,
                f'--username={username}',
                f'--password={password}',
                insecure_flag,
                _out=sys.stdout,
                _err=sys.stderr
            )
        except sh.ErrorReturnCode as error:
            raise StepRunnerException(f"Error logging in to ArgoCD: {error}") from error

    def __argocd_add_target_cluster(
        self,
        kube_api,
        kube_api_token=None,
        kube_api_skip_tls=False
    ):
        """If the target cluster is not the default cluster then add that cluster to ArgoCD.

        **WARNING:*** have not re-integration tested this since refactor. Have a feeling that
        it does not work as expected. I have a feeling there is issues with the context names.

        Raises
        ------
        StepRunnerException
            If error adding cluster to ArgoCD
        """
        # If the cluster is an external cluster and an api token was provided,
        # add the cluster to ArgoCD
        if kube_api != DEFAULT_CONFIG['kube-api-uri']:
            context_name = f'{kube_api}-context'
            kubeconfig = f"""---
apiVersion: v1
kind: Config
current-context: {context_name}
clusters:
- cluster:
    insecure-skip-tls-verify: {str(kube_api_skip_tls).lower()}
    server: {kube_api}
  name: default-cluster
contexts:
- context:
    cluster: default-cluster
    user: default-user
  name: {context_name}
preferences:
users:
- name: default-user
  user:
    token: {kube_api_token}
"""

            config_argocd_cluster_context_file = self.write_working_file(
                filename='config-argocd-cluster-context.yaml',
                contents=bytes(kubeconfig, 'utf-8')
            )
            try:
                sh.argocd.cluster.add(  # pylint: disable=no-member
                    '--kubeconfig', config_argocd_cluster_context_file,
                    context_name,
                    _out=sys.stdout,
                    _err=sys.stderr
                )
            except sh.ErrorReturnCode as error:
                raise StepRunnerException(
                    f"Error adding cluster ({kube_api}) to ArgoCD: {error}"
                ) from error

    @staticmethod
    def __argocd_app_create_or_update( # pylint: disable=too-many-arguments
        argocd_app_name,
        repo,
        revision,
        path,
        dest_server,
        auto_sync,
        values_files
    ):
        """Creates or updates an ArgoCD App.

        Raises
        ------
        StepRunnerException
            If error creating or updating ArgoCD app.
        """
        try:
            if str(auto_sync).lower() == 'true':
                sync_policy = 'automated'
            else:
                sync_policy = 'none'

            values_params = None
            if values_files:
                values_params = []
                for value_file in values_files:
                    values_params += [f'--values={value_file}']

            sh.argocd.app.create(  # pylint: disable=no-member
                argocd_app_name,
                f'--repo={repo}',
                f'--revision={revision}',
                f'--path={path}',
                f'--dest-server={dest_server}',
                f'--dest-namespace={argocd_app_name}',
                f'--sync-policy={sync_policy}',
                values_params,
                '--upsert',
                _out=sys.stdout,
                _err=sys.stderr
            )
        except sh.ErrorReturnCode as error:
            raise StepRunnerException(
                f"Error creating or updating ArgoCD app ({argocd_app_name}): {error}"
            ) from error

    @staticmethod
    def __argocd_app_sync(
        argocd_app_name,
        argocd_sync_timeout_seconds,
        argocd_sync_retry_limit
    ):
        try:
            sh.argocd.app.sync(  # pylint: disable=no-member
                '--prune',
                '--timeout', argocd_sync_timeout_seconds,
                '--retry-limit', argocd_sync_retry_limit,
                argocd_app_name,
                _out=sys.stdout,
                _err=sys.stderr
            )
        except sh.ErrorReturnCode as error:
            raise StepRunnerException(
                f"Error synchronization ArgoCD Application ({argocd_app_name}): {error}"
            ) from error

        try:
            sh.argocd.app.wait(  # pylint: disable=no-member
                '--timeout', argocd_sync_timeout_seconds,
                '--health',
                argocd_app_name,
                _out=sys.stdout,
                _err=sys.stderr
            )
        except sh.ErrorReturnCode as error:
            raise StepRunnerException(
                f"Error waiting for ArgoCD Application ({argocd_app_name}) synchronization: {error}"
            ) from error

    def __argocd_get_app_manifest(
        self,
        argocd_app_name,
        source='live'
    ):
        """Get ArgoCD Application manifest.

        Parameters
        ----------
        argocd_app_name : str
            Name of the ArgoCD Application to get the manifest for.
        source : str (live,git)
            Get the manifest from the 'live' version of the 'git' version.

        Returns
        -------
        str
            Path to the retrieved ArgoCD manifest file.

        Raises
        ------
        StepRunnerException
            If error getting ArgoCD manifest.
        """
        argocd_app_manifest_file = self.write_working_file('deploy_argocd_manifests.yml')
        try:
            sh.argocd.app.manifests(  # pylint: disable=no-member
                f'--source={source}',
                argocd_app_name,
                _out=argocd_app_manifest_file,
                _err=sys.stderr
            )
        except sh.ErrorReturnCode as error:
            raise StepRunnerException(
                f"Error reading ArgoCD Application ({argocd_app_name}) manifest: {error}"
            ) from error

        return argocd_app_manifest_file



import os
import re
from contextlib import redirect_stdout
from io import IOBase, StringIO
from unittest.mock import patch

import sh
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase
from tests.helpers.test_utils import Any, create_sh_side_effect
from ploigos_step_runner import StepResult
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.step_implementers.shared.openscap_generic import OpenSCAPGeneric


class BaseTestStepImplementerSharedOpenSCAPGeneric(BaseStepImplementerTestCase):
    def create_step_implementer(
            self,
            step_config={},
            step_name='',
            implementer='',
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=OpenSCAPGeneric,
            step_config=step_config,
            step_name=step_name,
            implementer=implementer,
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

class TestStepImplementerSharedOpenSCAPGeneric_step_implementer_config_defaults(
    BaseTestStepImplementerSharedOpenSCAPGeneric
):

    def test_result(self):
        defaults = OpenSCAPGeneric.step_implementer_config_defaults()
        expected_defaults = {
            'oscap-fetch-remote-resources': True,
            'container-image-pull-repository-type': 'containers-storage:',
            'container-image-repository-type': 'containers-storage:'
        }
        self.assertEqual(defaults, expected_defaults)

class TestStepImplementerSharedOpenSCAPGeneric__required_config_or_result_keys(
    BaseTestStepImplementerSharedOpenSCAPGeneric
):
    def test_result(self):
        required_keys = OpenSCAPGeneric._required_config_or_result_keys()
        expected_required_keys = [
            'oscap-input-definitions-uri',
            'container-image-tag',
            'container-image-pull-repository-type',
            ['container-image-pull-repository-type', 'container-image-repository-type']
        ]
        self.assertEqual(required_keys, expected_required_keys)

@patch("ploigos_step_runner.StepImplementer._validate_required_config_or_previous_step_result_artifact_keys")
class TestStepImplementerSharedOpenSCAPGeneric__validate_required_config_or_previous_step_result_artifact_keys(
    BaseTestStepImplementerSharedOpenSCAPGeneric
):
    def test__valid(self, mock_super_validate):
        step_config = {
            'oscap-input-definitions-uri': 'https://www.redhat.com/security/data/oval/v2/RHEL8/rhel-8.oval.xml.bz2',
            'image-tar-file': 'does-not-matter'
        }

        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='test',
                implementer='OpenSCAP',
                parent_work_dir_path=parent_work_dir_path
            )

            step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

            mock_super_validate.assert_called_once_with()

    def test_invalid_protocal(self, mock_super_validate):
        oscap_input_definitions_uri = 'foo://www.redhat.com/security/data/oval/v2/RHEL8/rhel-8.oval.xml.bz2'
        step_config = {
            'oscap-input-definitions-uri': oscap_input_definitions_uri,
            'image-tar-file': 'does-not-matter'
        }
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='test',
                implementer='OpenSCAP',
                parent_work_dir_path=parent_work_dir_path
            )

            with self.assertRaisesRegex(
                    AssertionError,
                    re.compile(
                        r'Open SCAP input definitions source '
                        rf'\({oscap_input_definitions_uri}\) must start with known protocol '
                        r'\(file://\|http://\|https://\)\.'
                    )
            ):
                step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

            mock_super_validate.assert_called_once_with()

    def test_invalid_extension(self, mock_super_validate):
        oscap_input_definitions_uri = 'https://www.redhat.com/security/data/oval/v2/RHEL8/rhel-8.oval.xml.foo'
        step_config = {
            'oscap-input-definitions-uri': oscap_input_definitions_uri,
            'image-tar-file': 'does-not-matter'
        }

        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='test',
                implementer='OpenSCAP',
                parent_work_dir_path=parent_work_dir_path,
            )

            with self.assertRaisesRegex(
                AssertionError,
                re.compile(
                    r'Open SCAP input definitions source '
                    rf'\({oscap_input_definitions_uri}\) must be of known type \(xml\|bz2\), got: \.foo'
                )
            ):
                step_implementer._validate_required_config_or_previous_step_result_artifact_keys()

            mock_super_validate.assert_called_once_with()

class TestStepImplementerSharedOpenSCAPGeneric___get_oscap_document_type(
    BaseTestStepImplementerSharedOpenSCAPGeneric
):
    @patch('sh.oscap', create=True)
    def test_sds(self, oscap_mock):
        oscap_input_file = '/does/not/matter.xml'

        sh.oscap.info.side_effect = create_sh_side_effect(
            mock_stdout="""
Document type: Source Data Stream
Imported: 2020-10-07T05:34:29

Stream: scap_org.open-scap_datastream_from_xccdf_com.redhat.rhsa-all.xml-xccdf12
Generated: (null)
Version: 1.2
Checklists:
	Ref-Id: scap_org.open-scap_cref_com.redhat.rhsa-all.xml-xccdf12
		Status: incomplete
		Resolved: true
		Profiles:
		Referenced check files:
			com.redhat.rhsa-all.xml
				system: http://oval.mitre.org/XMLSchema/oval-definitions-5
Checks:
	Ref-Id: scap_org.open-scap_cref_com.redhat.rhsa-all.xml
No dictionaries."""
        )

        oscap_document_type = OpenSCAPGeneric._OpenSCAPGeneric__get_oscap_document_type(
            oscap_input_file=oscap_input_file
        )

        sh.oscap.info.assert_called_once_with(
            oscap_input_file,
            _out=Any(IOBase)
        )

        self.assertEqual(oscap_document_type, 'Source Data Stream')

    @patch('sh.oscap', create=True)
    def test_oval(self, oscap_mock):
        oscap_input_file = '/does/not/matter.xml'

        sh.oscap.info.side_effect = create_sh_side_effect(
            mock_stdout="""
Document type: OVAL Definitions
OVAL version: 5.10
Generated: 2020-10-06T23:36:01
Imported: 2020-10-06T23:36:04"""
        )

        oscap_document_type = OpenSCAPGeneric._OpenSCAPGeneric__get_oscap_document_type(
            oscap_input_file=oscap_input_file
        )

        oscap_document_type = OpenSCAPGeneric._OpenSCAPGeneric__get_oscap_document_type(
            oscap_input_file=oscap_input_file
        )

        self.assertEqual(oscap_document_type, 'OVAL Definitions')

    @patch('sh.oscap', create=True)
    def test_xccdf(self, oscap_mock):
        oscap_input_file = '/does/not/matter.xml'

        sh.oscap.info.side_effect = create_sh_side_effect(
            mock_stdout="""
Document type: XCCDF Checklist
Checklist version: 1.1
Imported: 2020-02-11T13:41:07
Status: draft
Generated: 2020-02-11
Resolved: true
Profiles:
	Title: Protection Profile for General Purpose Operating Systems
		Id: ospp
	Title: PCI-DSS v3.2.1 Control Baseline for Red Hat Enterprise Linux 8
		Id: pci-dss
	Title: [DRAFT] DISA STIG for Red Hat Enterprise Linux 8
		Id: stig
	Title: Australian Cyber Security Centre (ACSC) Essential Eight
		Id: e8
Referenced check files:
	ssg-rhel8-oval.xml
		system: http://oval.mitre.org/XMLSchema/oval-definitions-5
	ssg-rhel8-ocil.xml
		system: http://scap.nist.gov/schema/ocil/2
	https://www.redhat.com/security/data/oval/com.redhat.rhsa-RHEL8.xml
		system: http://oval.mitre.org/XMLSchema/oval-definitions-5"""
        )

        oscap_document_type = OpenSCAPGeneric._OpenSCAPGeneric__get_oscap_document_type(
            oscap_input_file=oscap_input_file
        )

        sh.oscap.info.assert_called_once_with(
            oscap_input_file,
            _out=Any(IOBase)
        )

        self.assertEqual(oscap_document_type, 'XCCDF Checklist')

    @patch('sh.oscap', create=True)
    def test_error(self, oscap_mock):
        oscap_input_file = '/does/not/matter.xml'

        sh.oscap.info.side_effect = sh.ErrorReturnCode(
            'oscap info',
            b'mock out',
            b'mock error'
        )

        with self.assertRaisesRegex(
                StepRunnerException,
                re.compile(
                    r"Error getting document type of oscap"
                    rf" input file \({oscap_input_file}\):"
                    r".*RAN: oscap info"
                    r".*STDOUT:"
                    r".*mock out"
                    r".*STDERR:"
                    r".*mock error",
                    re.DOTALL
                )
        ):
            OpenSCAPGeneric._OpenSCAPGeneric__get_oscap_document_type(
                oscap_input_file=oscap_input_file
            )

        sh.oscap.info.assert_called_once_with(
            oscap_input_file,
            _out=Any(IOBase)
        )

class TestStepImplementerSharedOpenSCAPGeneric___get_oscap_eval_type_based_on_document_type(
    BaseTestStepImplementerSharedOpenSCAPGeneric
):
    def test_sds(self):
        oscap_document_type = 'Source Data Stream'

        oscap_eval_type = \
            OpenSCAPGeneric._OpenSCAPGeneric__get_oscap_eval_type_based_on_document_type(
                oscap_document_type=oscap_document_type
            )

        self.assertEqual(oscap_eval_type, 'xccdf')

    def test_xccdf(self):
        oscap_document_type = 'XCCDF Checklist'

        oscap_eval_type = \
            OpenSCAPGeneric._OpenSCAPGeneric__get_oscap_eval_type_based_on_document_type(
                oscap_document_type=oscap_document_type
            )

        self.assertEqual(oscap_eval_type, 'xccdf')

    def test_oval(self):
        oscap_document_type = 'OVAL Definitions'

        oscap_eval_type = \
            OpenSCAPGeneric._OpenSCAPGeneric__get_oscap_eval_type_based_on_document_type(
                oscap_document_type=oscap_document_type
            )

        self.assertEqual(oscap_eval_type, 'oval')

    def test_unknown(self):
        oscap_document_type = 'What is this nonsense?'

        oscap_eval_type = \
            OpenSCAPGeneric._OpenSCAPGeneric__get_oscap_eval_type_based_on_document_type(
                oscap_document_type=oscap_document_type
            )

        self.assertEqual(oscap_eval_type, None)

class TestStepImplementerSharedOpenSCAPGeneric___run_oscap_scan(
    BaseTestStepImplementerSharedOpenSCAPGeneric
):
    def __run_test_xccdf_do_not_fetch_remote_with_profile_all_pass(
            self,
            buildah_mock,
            oscap_eval_type,
            oscap_fetch_remote_resources,
            oscap_stdout,
            oscap_stdout_expected,
            oscap_profile=None,
            oscap_tailoring_file=None,
            oscap_eval_success_expected=True,
            exit_code=0,
            oscap_eval_fails_expected=None
    ):
        with TempDirectory() as temp_dir:
            buildah_unshare_command = sh.buildah.bake('unshare')
            oscap_input_file = '/does/not/matter/input.xml'
            oscap_out_file_path = os.path.join(temp_dir.path, 'out')
            oscap_xml_results_file_path = '/does/not/matter/results.xml'
            oscap_html_report_path = '/does/not/matter/results.html'
            container_mount_path = '/does/not/matter/coutainer_mount'

            exception = None
            if exit_code == 2:
                exception = sh.ErrorReturnCode_2(
                    'oscap-chroot eval',
                    bytes(oscap_stdout, 'utf-8'),
                    bytes(f'mock error - exit code {exit_code}', 'utf-8')
                )
            elif exit_code == 1:
                exception = sh.ErrorReturnCode_1(
                    'oscap-chroot eval',
                    bytes(oscap_stdout, 'utf-8'),
                    bytes(f'mock error - exit code {exit_code}', 'utf-8')
                )
            elif exit_code:
                exception = sh.ErrorReturnCode(
                    'oscap-chroot eval',
                    bytes(oscap_stdout, 'utf-8'),
                    bytes(f'mock error - exit code {exit_code}', 'utf-8')
                )

            buildah_mock.bake('unshare').bake('oscap-chroot').side_effect = create_sh_side_effect(
                mock_stdout=oscap_stdout,
                exception=exception
            )

            stdout_buff = StringIO()
            with redirect_stdout(stdout_buff):
                oscap_eval_success, oscap_eval_fails = OpenSCAPGeneric._OpenSCAPGeneric__run_oscap_scan(
                    buildah_unshare_command=buildah_unshare_command,
                    oscap_eval_type=oscap_eval_type,
                    oscap_input_file=oscap_input_file,
                    oscap_out_file_path=oscap_out_file_path,
                    oscap_xml_results_file_path=oscap_xml_results_file_path,
                    oscap_html_report_path=oscap_html_report_path,
                    container_mount_path=container_mount_path,
                    oscap_profile=oscap_profile,
                    oscap_tailoring_file=oscap_tailoring_file,
                    oscap_fetch_remote_resources=oscap_fetch_remote_resources
                )

            if oscap_profile:
                oscap_profile_flag = f"--profile={oscap_profile}"
            else:
                oscap_profile_flag = None

            if oscap_fetch_remote_resources:
                oscap_fetch_remote_resources_flag = "--fetch-remote-resources"
            else:
                oscap_fetch_remote_resources_flag = None

            if oscap_tailoring_file:
                oscap_tailoring_file_flag = f"--tailoring-file={oscap_tailoring_file}"
            else:
                oscap_tailoring_file_flag = None

            buildah_mock.bake('unshare').bake.assert_called_with('oscap-chroot')
            buildah_mock.bake('unshare').bake('oscap-chroot').assert_called_once_with(
                container_mount_path,
                oscap_eval_type,
                'eval',
                oscap_profile_flag,
                oscap_fetch_remote_resources_flag,
                oscap_tailoring_file_flag,
                f'--results={oscap_xml_results_file_path}',
                f'--report={oscap_html_report_path}',
                oscap_input_file,
                _out=Any(IOBase),
                _err=Any(IOBase),
                _tee='err'
            )

            self.assertEqual(oscap_eval_success, oscap_eval_success_expected)

            if oscap_eval_fails_expected:
                self.assertEqual(oscap_eval_fails, oscap_eval_fails_expected)

            stdout = stdout_buff.getvalue()
            self.assertEqual(stdout, oscap_stdout_expected)

    @patch('sh.buildah', create=True)
    def test_xccdf_do_no_fetch_remote_with_profile_all_pass(self, buildah_mock):
        self.__run_test_xccdf_do_not_fetch_remote_with_profile_all_pass(
            buildah_mock=buildah_mock,
            oscap_eval_type='xccdf',
            oscap_profile='this.is.real.i.sware',
            oscap_fetch_remote_resources=False,
            oscap_stdout="""
Title\r	Enable Kernel Page-Table Isolation (KPTI)
Rule\r	xccdf_org.ssgproject.content_rule_grub2_pti_argument
Ident\r	CCE-82194-2
Result\r	pass

Title\r	Install dnf-automatic Package
Rule\r	xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed
Ident\r	CCE-82985-3
Result\r	pass""",
            oscap_stdout_expected="""
Title	Enable Kernel Page-Table Isolation (KPTI)
Rule	xccdf_org.ssgproject.content_rule_grub2_pti_argument
Ident	CCE-82194-2
Result	pass

Title	Install dnf-automatic Package
Rule	xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed
Ident	CCE-82985-3
Result	pass
"""
        )

    @patch('sh.buildah', create=True)
    def test_xccdf_do_yes_fetch_remote_with_profile_all_pass(self, buildah_mock):
        self.__run_test_xccdf_do_not_fetch_remote_with_profile_all_pass(
            buildah_mock=buildah_mock,
            oscap_eval_type='xccdf',
            oscap_profile='this.is.real.i.sware',
            oscap_fetch_remote_resources=True,
            oscap_stdout="""
Title\r	Enable Kernel Page-Table Isolation (KPTI)
Rule\r	xccdf_org.ssgproject.content_rule_grub2_pti_argument
Ident\r	CCE-82194-2
Result\r	pass

Title\r	Install dnf-automatic Package
Rule\r	xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed
Ident\r	CCE-82985-3
Result\r	pass""",
            oscap_stdout_expected="""
Title	Enable Kernel Page-Table Isolation (KPTI)
Rule	xccdf_org.ssgproject.content_rule_grub2_pti_argument
Ident	CCE-82194-2
Result	pass

Title	Install dnf-automatic Package
Rule	xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed
Ident	CCE-82985-3
Result	pass
"""
        )

    @patch('sh.buildah', create=True)
    def test_xccdf_do_yes_fetch_remote_no_profile_all_pass(self, buildah_mock):
        self.__run_test_xccdf_do_not_fetch_remote_with_profile_all_pass(
            buildah_mock=buildah_mock,
            oscap_eval_type='xccdf',
            oscap_profile=None,
            oscap_fetch_remote_resources=True,
            oscap_stdout="""
Title\r	Enable Kernel Page-Table Isolation (KPTI)
Rule\r	xccdf_org.ssgproject.content_rule_grub2_pti_argument
Ident\r	CCE-82194-2
Result\r	pass

Title\r	Install dnf-automatic Package
Rule\r	xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed
Ident\r	CCE-82985-3
Result\r	pass""",
            oscap_stdout_expected="""
Title	Enable Kernel Page-Table Isolation (KPTI)
Rule	xccdf_org.ssgproject.content_rule_grub2_pti_argument
Ident	CCE-82194-2
Result	pass

Title	Install dnf-automatic Package
Rule	xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed
Ident	CCE-82985-3
Result	pass
"""
        )

    @patch('sh.buildah', create=True)
    def test_xccdf_do_yes_fetch_remote_with_profile_with_fail(self, buildah_mock):
        self.__run_test_xccdf_do_not_fetch_remote_with_profile_all_pass(
            buildah_mock=buildah_mock,
            oscap_eval_type='xccdf',
            oscap_profile='this.is.real.i.sware',
            oscap_fetch_remote_resources=True,
            oscap_eval_success_expected=False,
            exit_code=2,
            oscap_stdout="""
Title\r	Enable Kernel Page-Table Isolation (KPTI)
Rule\r	xccdf_org.ssgproject.content_rule_grub2_pti_argument
Ident\r	CCE-82194-2
Result\r	notapplicable

Title\r	Install dnf-automatic Package
Rule\r	xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed
Ident\r	CCE-82985-3
Result\r	fail

Title\r	Ensure gpgcheck Enabled for All yum Package Repositories
Rule\r	xccdf_org.ssgproject.content_rule_ensure_gpgcheck_never_disabled
Ident\r	CCE-80792-5
Result\r	pass""",
            oscap_stdout_expected="""
Title	Enable Kernel Page-Table Isolation (KPTI)
Rule	xccdf_org.ssgproject.content_rule_grub2_pti_argument
Ident	CCE-82194-2
Result	notapplicable

Title	Install dnf-automatic Package
Rule	xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed
Ident	CCE-82985-3
Result	fail

Title	Ensure gpgcheck Enabled for All yum Package Repositories
Rule	xccdf_org.ssgproject.content_rule_ensure_gpgcheck_never_disabled
Ident	CCE-80792-5
Result	pass
""",
            oscap_eval_fails_expected="""
Title	Install dnf-automatic Package
Rule	xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed
Ident	CCE-82985-3
Result	fail
"""
        )

    @patch('sh.buildah', create=True)
    def test_str_oscap_fetch_remote_resources_flag(self, buildah_mock):
        self.__run_test_xccdf_do_not_fetch_remote_with_profile_all_pass(
            buildah_mock=buildah_mock,
            oscap_eval_type='xccdf',
            oscap_profile='this.is.real.i.sware',
            oscap_fetch_remote_resources="True",
            oscap_stdout="""
Title\r	Enable Kernel Page-Table Isolation (KPTI)
Rule\r	xccdf_org.ssgproject.content_rule_grub2_pti_argument
Ident\r	CCE-82194-2
Result\r	pass

Title\r	Install dnf-automatic Package
Rule\r	xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed
Ident\r	CCE-82985-3
Result\r	pass""",
            oscap_stdout_expected="""
Title	Enable Kernel Page-Table Isolation (KPTI)
Rule	xccdf_org.ssgproject.content_rule_grub2_pti_argument
Ident	CCE-82194-2
Result	pass

Title	Install dnf-automatic Package
Rule	xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed
Ident	CCE-82985-3
Result	pass
"""
        )

    @patch('sh.buildah', create=True)
    def test_oval_do_no_fetch_remote_with_profile_all_pass(self, buildah_mock):
        self.__run_test_xccdf_do_not_fetch_remote_with_profile_all_pass(
            buildah_mock=buildah_mock,
            oscap_eval_type='oval',
            oscap_fetch_remote_resources=False,
            oscap_stdout="""
Definition oval:com.redhat.rhsa:def:20203699: false
Definition oval:com.redhat.rhsa:def:20203669: false
Definition oval:com.redhat.rhsa:def:20203665: false
Definition oval:com.redhat.rhsa:def:20203662: false""",
            oscap_stdout_expected="""
Definition oval:com.redhat.rhsa:def:20203699: false
Definition oval:com.redhat.rhsa:def:20203669: false
Definition oval:com.redhat.rhsa:def:20203665: false
Definition oval:com.redhat.rhsa:def:20203662: false
"""
        )

    @patch('sh.buildah', create=True)
    def test_oval_do_no_fetch_remote_with_profile_all_pass(self, buildah_mock):
        self.__run_test_xccdf_do_not_fetch_remote_with_profile_all_pass(
            buildah_mock=buildah_mock,
            oscap_eval_type='oval',
            oscap_fetch_remote_resources=False,
            oscap_eval_success_expected=False,
            oscap_stdout="""
Definition oval:com.redhat.rhsa:def:20203699: false
Definition oval:com.redhat.rhsa:def:20203669: true
Definition oval:com.redhat.rhsa:def:20203665: false
Definition oval:com.redhat.rhsa:def:20203662: true""",
            oscap_stdout_expected="""
Definition oval:com.redhat.rhsa:def:20203699: false
Definition oval:com.redhat.rhsa:def:20203669: true
Definition oval:com.redhat.rhsa:def:20203665: false
Definition oval:com.redhat.rhsa:def:20203662: true
""",
            oscap_eval_fails_expected="""Definition oval:com.redhat.rhsa:def:20203669: true
Definition oval:com.redhat.rhsa:def:20203662: true
"""
        )

    @patch('sh.buildah', create=True)
    def test_xccdf_exit_code_1(self, buildah_mock):
        with self.assertRaisesRegex(
                StepRunnerException,
                re.compile(
                    r"Error running 'oscap xccdf eval':"
                    r".*RAN: oscap-chroot eval"
                    r".*STDOUT:"
                    r".*Title\s*Enable Kernel Page-Table Isolation \(KPTI\)"
                    r"\s*Rule\s*xccdf_org.ssgproject.content_rule_grub2_pti_argument"
                    r"\s*Ident\s*CCE-82194-2"
                    r"\s*Result\s*notapplicable"
                    r".*Title\s*Install dnf-automatic Package"
                    r"\s*Rule\s*xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed"
                    r"\s*Ident\s*CCE-82985-3"
                    r"\s*Result\s*fail"
                    r".*Title\s*Ensure gpgcheck Enabled for All yum Package Repositories"
                    r"\s*Rule\s*xccdf_org.ssgproject.content_rule_ensure_gpgcheck_never_disabled"
                    r"\s*Ident\s*CCE-80792-5"
                    r"\s*Result\s*pass"
                    r".*STDERR:"
                    r".*mock error - exit code 1",
                    re.DOTALL
                )
        ):
            self.__run_test_xccdf_do_not_fetch_remote_with_profile_all_pass(
                buildah_mock=buildah_mock,
                oscap_eval_type='xccdf',
                oscap_profile='this.is.real.i.sware',
                oscap_fetch_remote_resources=True,
                oscap_eval_success_expected=False,
                exit_code=1,
                oscap_stdout="""
    Title\r	Enable Kernel Page-Table Isolation (KPTI)
    Rule\r	xccdf_org.ssgproject.content_rule_grub2_pti_argument
    Ident\r	CCE-82194-2
    Result\r	notapplicable

    Title\r	Install dnf-automatic Package
    Rule\r	xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed
    Ident\r	CCE-82985-3
    Result\r	fail

    Title\r	Ensure gpgcheck Enabled for All yum Package Repositories
    Rule\r	xccdf_org.ssgproject.content_rule_ensure_gpgcheck_never_disabled
    Ident\r	CCE-80792-5
    Result\r	pass""",
                oscap_stdout_expected="""
    Title	Enable Kernel Page-Table Isolation (KPTI)
    Rule	xccdf_org.ssgproject.content_rule_grub2_pti_argument
    Ident	CCE-82194-2
    Result	notapplicable

    Title	Install dnf-automatic Package
    Rule	xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed
    Ident	CCE-82985-3
    Result	fail

    Title	Ensure gpgcheck Enabled for All yum Package Repositories
    Rule	xccdf_org.ssgproject.content_rule_ensure_gpgcheck_never_disabled
    Ident	CCE-80792-5
    Result	pass
    """
            )

    @patch('sh.buildah', create=True)
    def test_oval_exit_code_1(self, buildah_mock):
        with self.assertRaisesRegex(
                StepRunnerException,
                re.compile(
                    r"Error running 'oscap oval eval': "
                    r".*RAN: oscap-chroot eval"
                    r".*STDOUT:"
                    r".*Definition oval:com.redhat.rhsa:def:20203699: false"
                    r".*Definition oval:com.redhat.rhsa:def:20203669: true"
                    r".*Definition oval:com.redhat.rhsa:def:20203665: false"
                    r".*Definition oval:com.redhat.rhsa:def:20203662: true"
                    r".*STDERR:"
                    r".*mock error - exit code 1",
                    re.DOTALL
                )
        ):
            self.__run_test_xccdf_do_not_fetch_remote_with_profile_all_pass(
                buildah_mock=buildah_mock,
                oscap_eval_type='oval',
                oscap_fetch_remote_resources=False,
                oscap_eval_success_expected=False,
                exit_code=1,
                oscap_stdout="""
    Definition oval:com.redhat.rhsa:def:20203699: false
    Definition oval:com.redhat.rhsa:def:20203669: true
    Definition oval:com.redhat.rhsa:def:20203665: false
    Definition oval:com.redhat.rhsa:def:20203662: true""",
                oscap_stdout_expected="""
    Definition oval:com.redhat.rhsa:def:20203699: false
    Definition oval:com.redhat.rhsa:def:20203669: true
    Definition oval:com.redhat.rhsa:def:20203665: false
    Definition oval:com.redhat.rhsa:def:20203662: true
    """
            )

    @patch('sh.buildah', create=True)
    def test_oval_exit_code_2(self, buildah_mock):
        with self.assertRaisesRegex(
                StepRunnerException,
                re.compile(
                    r"Error running 'oscap oval eval': "
                    r".*RAN: oscap-chroot eval"
                    r".*STDOUT:"
                    r".*Definition oval:com.redhat.rhsa:def:20203699: false"
                    r".*Definition oval:com.redhat.rhsa:def:20203669: true"
                    r".*Definition oval:com.redhat.rhsa:def:20203665: false"
                    r".*Definition oval:com.redhat.rhsa:def:20203662: true"
                    r".*STDERR:"
                    r".*mock error - exit code 2",
                    re.DOTALL
                )
        ):
            self.__run_test_xccdf_do_not_fetch_remote_with_profile_all_pass(
                buildah_mock=buildah_mock,
                oscap_eval_type='oval',
                oscap_fetch_remote_resources=False,
                oscap_eval_success_expected=False,
                exit_code=2,
                oscap_stdout="""
    Definition oval:com.redhat.rhsa:def:20203699: false
    Definition oval:com.redhat.rhsa:def:20203669: true
    Definition oval:com.redhat.rhsa:def:20203665: false
    Definition oval:com.redhat.rhsa:def:20203662: true""",
                oscap_stdout_expected="""
    Definition oval:com.redhat.rhsa:def:20203699: false
    Definition oval:com.redhat.rhsa:def:20203669: true
    Definition oval:com.redhat.rhsa:def:20203665: false
    Definition oval:com.redhat.rhsa:def:20203662: true
    """
            )

    @patch('sh.buildah', create=True)
    def test_oval_exit_code_unknown(self, buildah_mock):
        with self.assertRaisesRegex(
                StepRunnerException,
                re.compile(
                    r"Error running 'oscap oval eval': "
                    r".*RAN: oscap-chroot eval"
                    r".*STDOUT:"
                    r".*Definition oval:com.redhat.rhsa:def:20203699: false"
                    r".*Definition oval:com.redhat.rhsa:def:20203669: true"
                    r".*Definition oval:com.redhat.rhsa:def:20203665: false"
                    r".*Definition oval:com.redhat.rhsa:def:20203662: true"
                    r".*STDERR:"
                    r".*mock error - exit code 42",
                    re.DOTALL
                )
        ):
            self.__run_test_xccdf_do_not_fetch_remote_with_profile_all_pass(
                buildah_mock=buildah_mock,
                oscap_eval_type='oval',
                oscap_fetch_remote_resources=False,
                oscap_eval_success_expected=False,
                exit_code=42,
                oscap_stdout="""
    Definition oval:com.redhat.rhsa:def:20203699: false
    Definition oval:com.redhat.rhsa:def:20203669: true
    Definition oval:com.redhat.rhsa:def:20203665: false
    Definition oval:com.redhat.rhsa:def:20203662: true""",
                oscap_stdout_expected="""
    Definition oval:com.redhat.rhsa:def:20203699: false
    Definition oval:com.redhat.rhsa:def:20203669: true
    Definition oval:com.redhat.rhsa:def:20203665: false
    Definition oval:com.redhat.rhsa:def:20203662: true
    """
            )

    @patch('sh.buildah', create=True)
    def test_xccdf_with_tailoring_file(self, buildah_mock):
        self.__run_test_xccdf_do_not_fetch_remote_with_profile_all_pass(
            buildah_mock=buildah_mock,
            oscap_eval_type='xccdf',
            oscap_profile='this.is.real.i.sware',
            oscap_fetch_remote_resources=False,
            oscap_tailoring_file="/does/not/matter/tailoring.xccdf.xml",
            oscap_stdout="""
Title\r	Enable Kernel Page-Table Isolation (KPTI)
Rule\r	xccdf_org.ssgproject.content_rule_grub2_pti_argument
Ident\r	CCE-82194-2
Result\r	pass

Title\r	Install dnf-automatic Package
Rule\r	xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed
Ident\r	CCE-82985-3
Result\r	pass""",
            oscap_stdout_expected="""
Title	Enable Kernel Page-Table Isolation (KPTI)
Rule	xccdf_org.ssgproject.content_rule_grub2_pti_argument
Ident	CCE-82194-2
Result	pass

Title	Install dnf-automatic Package
Rule	xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed
Ident	CCE-82985-3
Result	pass
"""
        )

class TestStepImplementerSharedOpenSCAPGeneric__run_step(
    BaseTestStepImplementerSharedOpenSCAPGeneric
):
    @patch.object(OpenSCAPGeneric, '_OpenSCAPGeneric__run_oscap_scan')
    @patch('ploigos_step_runner.step_implementers.shared.openscap_generic.mount_container')
    @patch(
        'ploigos_step_runner.step_implementers.shared.openscap_generic.create_container_from_image',
        return_value='mock-image-working-container-mock-1'
    )
    @patch.object(OpenSCAPGeneric, '_OpenSCAPGeneric__get_oscap_document_type')
    @patch('sh.buildah', create=True)
    def test_pass(
        self,
        buildah_mock,
        get_oscap_document_type_mock,
        create_container_from_image_mock,
        mount_container_mock,
        run_oscap_scan_mock
    ):
        oscap_document_type = 'Source Data Stream'
        oscap_eval_type = 'xccdf'
        oscap_input_definitions_uri = 'https://www.redhat.com/security/data/metrics/ds/v2/RHEL8/rhel-8.ds.xml.bz2'
        container_image_tag = 'localhost/mock-org/mock-image:v0.42.0-mock'
        step_config = {
            'oscap-input-definitions-uri': oscap_input_definitions_uri,
            'oscap-profile': 'foo',
            'container-image-tag': container_image_tag
        }
        oscap_eval_success = True
        oscap_eval_fails = None

        with TempDirectory() as temp_dir:
            # setup test
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            mount_path = '/does/not/matter/container-mount'

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='test',
                implementer='OpenSCAP',
                parent_work_dir_path=parent_work_dir_path
            )

            get_oscap_document_type_mock.return_value = oscap_document_type
            mount_container_mock.return_value = mount_path
            run_oscap_scan_mock.return_value = [
                oscap_eval_success,
                oscap_eval_fails
            ]

            # run test
            stdout_buff = StringIO()
            with redirect_stdout(stdout_buff):
                step_result = step_implementer._run_step()

            # verify results
            expected_results = StepResult(
                step_name='test',
                sub_step_name='OpenSCAP',
                sub_step_implementer_name='OpenSCAP'
            )
            expected_results.success=True
            expected_results.add_artifact(
                name='html-report',
                value=f"{step_implementer.work_dir_path}/oscap-xccdf-report.html"
            )
            expected_results.add_artifact(
                name='xml-report',
                value=f"{step_implementer.work_dir_path}/oscap-xccdf-results.xml"
            )
            expected_results.add_artifact(
                name='stdout-report',
                value=f"{step_implementer.work_dir_path}/oscap-xccdf-out"
            )
            self.assertEqual(expected_results, step_result)

            stdout = stdout_buff.getvalue()
            expected_container_name = 'mock-image-working-container-mock-1'
            self.assertRegex(
                stdout,
                re.compile(
                    rf".*Create container from image \({container_image_tag}\)"
                    rf".*Created container \({expected_container_name}\) from image \({container_image_tag}\)"
                    rf".*Mount container: {expected_container_name}"
                    rf".*Mounted container \({expected_container_name}\) with mount path: '{mount_path}'"
                    rf".*Download input definitions: {oscap_input_definitions_uri}"
                    rf".*Downloaded input definitions to: /.+/working/test/rhel\-8.ds.xml"
                    rf".*Determine OpenSCAP document type of input file: /.+/working/test/rhel\-8\.ds\.xml"
                    rf".*Determined OpenSCAP document type of input file \(/.+/working/test/rhel\-8\.ds\.xml\): {oscap_document_type}"
                    rf".*Determine OpenSCAP eval type for input file \(/.+/working/test/rhel\-8\.ds\.xml\) of document type: {oscap_document_type}"
                    rf".*Determined OpenSCAP eval type of input file \(/.+/working/test/rhel\-8\.ds\.xml\): {oscap_eval_type}"
                    rf".*Run oscap scan"
                    rf".*OpenSCAP scan completed with eval success",
                    re.DOTALL
                )
            )

    @patch.object(OpenSCAPGeneric, '_OpenSCAPGeneric__run_oscap_scan')
    @patch('ploigos_step_runner.step_implementers.shared.openscap_generic.mount_container')
    @patch(
        'ploigos_step_runner.step_implementers.shared.openscap_generic.create_container_from_image',
        return_value='mock-image-working-container-mock-1'
    )
    @patch.object(OpenSCAPGeneric, '_OpenSCAPGeneric__get_oscap_document_type')
    @patch('sh.buildah', create=True)
    def test_fail(
        self,
        buildah_mock,
        get_oscap_document_type_mock,
        create_container_from_image_mock,
        mount_container_mock,
        run_oscap_scan_mock
    ):
        oscap_document_type = 'Source Data Stream'
        oscap_eval_type = 'xccdf'
        oscap_input_definitions_uri = 'https://www.redhat.com/security/data/metrics/ds/v2/RHEL8/rhel-8.ds.xml.bz2'
        container_image_tag = 'localhost/mock-org/mock-image:v0.42.0-mock'
        step_config = {
            'oscap-input-definitions-uri': oscap_input_definitions_uri,
            'oscap-profile': 'foo',
            'container-image-tag': container_image_tag
        }
        oscap_eval_success = False
        oscap_eval_fails = """
Title	Install dnf-automatic Package
Rule	xccdf_org.ssgproject.content_rule_package_dnf-automatic_installed
Ident	CCE-82985-3
Result	fail
"""

        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            mount_path = '/does/not/matter/container-mount'

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='test',
                implementer='OpenSCAP',
                parent_work_dir_path=parent_work_dir_path
            )

            get_oscap_document_type_mock.return_value = oscap_document_type
            mount_container_mock.return_value = mount_path
            run_oscap_scan_mock.return_value = [
                oscap_eval_success,
                oscap_eval_fails
            ]

            stdout_buff = StringIO()
            with redirect_stdout(stdout_buff):
                step_result = step_implementer._run_step()

            expected_results = StepResult(
                step_name='test',
                sub_step_name='OpenSCAP',
                sub_step_implementer_name='OpenSCAP'
            )
            expected_results.success = False
            expected_results.message = "OSCAP eval found issues:\n\n" \
                "Title\tInstall dnf-automatic Package\n" \
                "Rule\txccdf_org.ssgproject.content_rule_package_dnf-automatic_installed" \
                "\nIdent\tCCE-82985-3\nResult\tfail\n"
            expected_results.add_artifact(
                name='html-report',
                value=f"{step_implementer.work_dir_path}/oscap-xccdf-report.html"
            )
            expected_results.add_artifact(
                name='xml-report',
                value=f"{step_implementer.work_dir_path}/oscap-xccdf-results.xml"
            )
            expected_results.add_artifact(
                name='stdout-report',
                value=f"{step_implementer.work_dir_path}/oscap-xccdf-out"
            )
            self.assertEqual(expected_results, step_result)

            stdout = stdout_buff.getvalue()
            expected_container_name = 'mock-image-working-container-mock-1'
            self.assertRegex(
                stdout,
                re.compile(
                    rf".*Create container from image \({container_image_tag}\)"
                    rf".*Created container \({expected_container_name}\) from image \({container_image_tag}\)"
                    rf".*Mount container: {expected_container_name}"
                    rf".*Mounted container \({expected_container_name}\) with mount path: '{mount_path}'"
                    rf".*Download input definitions: {oscap_input_definitions_uri}"
                    rf".*Downloaded input definitions to: /.+/working/test/rhel\-8.ds.xml"
                    rf".*Determine OpenSCAP document type of input file: /.+/working/test/rhel\-8\.ds\.xml"
                    rf".*Determined OpenSCAP document type of input file \(/.+/working/test/rhel\-8\.ds\.xml\): {oscap_document_type}"
                    rf".*Determine OpenSCAP eval type for input file \(/.+/working/test/rhel\-8\.ds\.xml\) of document type: {oscap_document_type}"
                    rf".*Determined OpenSCAP eval type of input file \(/.+/working/test/rhel\-8\.ds\.xml\): {oscap_eval_type}"
                    rf".*Run oscap scan"
                    rf".*OpenSCAP scan completed with eval success: False",
                    re.DOTALL
                )
            )

    @patch.object(OpenSCAPGeneric, '_OpenSCAPGeneric__run_oscap_scan')
    @patch('ploigos_step_runner.step_implementers.shared.openscap_generic.mount_container')
    @patch(
        'ploigos_step_runner.step_implementers.shared.openscap_generic.create_container_from_image',
        return_value='mock-image-working-container-mock-1'
    )
    @patch.object(OpenSCAPGeneric, '_OpenSCAPGeneric__get_oscap_document_type')
    @patch('sh.buildah', create=True)
    def test_pass_with_tailoring_file(
        self,
        buildah_mock,
        get_oscap_document_type_mock,
        create_container_from_image_mock,
        mount_container_mock,
        run_oscap_scan_mock
    ):
        oscap_document_type = 'Source Data Stream'
        oscap_eval_type = 'xccdf'
        oscap_input_definitions_uri = 'https://www.redhat.com/security/data/metrics/ds/v2/RHEL8/rhel-8.ds.xml.bz2'
        container_image_tag = 'localhost/mock-org/mock-image:v0.42.0-mock'
        oscap_tailoring_uri = 'https://raw.githubusercontent.com/ploigos/ploigos-example-oscap-content/main/xccdf_com.redhat.ploigos_profile_example_ubi8-tailoring-xccdf.xml'
        step_config = {
            'oscap-input-definitions-uri': oscap_input_definitions_uri,
            'oscap-tailoring-uri': oscap_tailoring_uri,
            'oscap-profile': 'foo',
            'container-image-tag': container_image_tag
        }
        oscap_eval_success = True
        oscap_eval_fails = None

        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            mount_path = '/does/not/matter/container-mount'

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='test',
                implementer='OpenSCAP',
                parent_work_dir_path=parent_work_dir_path
            )

            get_oscap_document_type_mock.return_value = oscap_document_type
            mount_container_mock.return_value = mount_path
            run_oscap_scan_mock.return_value = [
                oscap_eval_success,
                oscap_eval_fails
            ]

            stdout_buff = StringIO()
            with redirect_stdout(stdout_buff):
                step_result = step_implementer._run_step()

            expected_results = StepResult(
                step_name='test',
                sub_step_name='OpenSCAP',
                sub_step_implementer_name='OpenSCAP'
            )
            expected_results.success=True
            expected_results.add_artifact(
                name='html-report',
                value=f"{step_implementer.work_dir_path}/oscap-xccdf-report.html"
            )
            expected_results.add_artifact(
                name='xml-report',
                value=f"{step_implementer.work_dir_path}/oscap-xccdf-results.xml"
            )
            expected_results.add_artifact(
                name='stdout-report',
                value=f"{step_implementer.work_dir_path}/oscap-xccdf-out"
            )
            self.assertEqual(expected_results, step_result)

            stdout = stdout_buff.getvalue()

            expected_container_name = 'mock-image-working-container-mock-1'
            self.assertRegex(
                stdout,
                re.compile(
                    rf".*Create container from image \({container_image_tag}\)"
                    rf".*Created container \({expected_container_name}\) from image \({container_image_tag}\)"
                    rf".*Mount container: {expected_container_name}"
                    rf".*Mounted container \({expected_container_name}\) with mount path: '{mount_path}'"
                    rf".*Download input definitions: {oscap_input_definitions_uri}"
                    rf".*Downloaded input definitions to: /.+/working/test/rhel\-8.ds.xml"
                    rf".*Download oscap tailoring file: {oscap_tailoring_uri}"
                    rf".*Download oscap tailoring file to: /.+/working/test/xccdf_com.redhat\.ploigos_profile_example_ubi8\-tailoring\-xccdf\.xml"
                    rf".*Determine OpenSCAP document type of input file: /.+/working/test/rhel\-8\.ds\.xml"
                    rf".*Determined OpenSCAP document type of input file \(/.+/working/test/rhel\-8\.ds\.xml\): {oscap_document_type}"
                    rf".*Determine OpenSCAP eval type for input file \(/.+/working/test/rhel\-8\.ds\.xml\) of document type: {oscap_document_type}"
                    rf".*Determined OpenSCAP eval type of input file \(/.+/working/test/rhel\-8\.ds\.xml\): {oscap_eval_type}"
                    rf".*Run oscap scan"
                    rf".*OpenSCAP scan completed with eval success",
                    re.DOTALL
                )
            )

    @patch('ploigos_step_runner.step_implementers.shared.openscap_generic.download_and_decompress_source_to_destination')
    @patch.object(OpenSCAPGeneric, '_OpenSCAPGeneric__run_oscap_scan')
    @patch('ploigos_step_runner.step_implementers.shared.openscap_generic.mount_container')
    @patch(
        'ploigos_step_runner.step_implementers.shared.openscap_generic.create_container_from_image',
        return_value='mock-image-working-container-mock-1'
    )
    @patch.object(OpenSCAPGeneric, '_OpenSCAPGeneric__get_oscap_document_type')
    @patch('sh.buildah', create=True)
    def test_fail_downloading_open_scap_input_file(
        self,
        buildah_mock,
        get_oscap_document_type_mock,
        create_container_from_image_mock,
        mount_container_mock,
        run_oscap_scan_mock,
        download_and_decompress_source_to_destination_mock
    ):
        oscap_document_type = 'Source Data Stream'
        oscap_eval_type = 'xccdf'
        oscap_input_definitions_uri = 'https://www.redhat.com/security/data/metrics/ds/v2/RHEL8/rhel-8.ds.xml.bz2'
        container_image_tag = 'localhost/mock-org/mock-image:v0.42.0-mock'
        step_config = {
            'oscap-input-definitions-uri': oscap_input_definitions_uri,
            'oscap-profile': 'foo',
            'container-image-tag': container_image_tag
        }
        oscap_eval_success = True
        oscap_eval_fails = None

        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            mount_path = '/does/not/matter/container-mount'

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='test',
                implementer='OpenSCAP',
                parent_work_dir_path=parent_work_dir_path
            )

            get_oscap_document_type_mock.return_value = oscap_document_type
            mount_container_mock.return_value = mount_path
            run_oscap_scan_mock.return_value = [
                oscap_eval_success,
                oscap_eval_fails
            ]

            mock_error_msg = 'mock error downloading open scap file'
            download_and_decompress_source_to_destination_mock.side_effect = RuntimeError(
                mock_error_msg
            )

            # run test
            stdout_buff = StringIO()
            with redirect_stdout(stdout_buff):
                step_result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='test',
                sub_step_name='OpenSCAP',
                sub_step_implementer_name='OpenSCAP'
            )
            expected_step_result.success = False
            expected_step_result.message = 'Error downloading OpenSCAP input file: ' \
                f'{mock_error_msg}'
            self.assertEqual(expected_step_result, step_result)

            stdout = stdout_buff.getvalue()
            expected_container_name = 'mock-image-working-container-mock-1'
            self.assertRegex(
                stdout,
                re.compile(
                    rf".*Create container from image \({container_image_tag}\)"
                    rf".*Created container \({expected_container_name}\) from image \({container_image_tag}\)"
                    rf".*Mount container: {expected_container_name}"
                    rf".*Mounted container \({expected_container_name}\) with mount path: '{mount_path}'"
                    rf".*Download input definitions: {oscap_input_definitions_uri}",
                    re.DOTALL
                )
            )

    @patch('ploigos_step_runner.step_implementers.shared.openscap_generic.download_and_decompress_source_to_destination')
    @patch.object(OpenSCAPGeneric, '_OpenSCAPGeneric__run_oscap_scan')
    @patch('ploigos_step_runner.step_implementers.shared.openscap_generic.mount_container')
    @patch(
        'ploigos_step_runner.step_implementers.shared.openscap_generic.create_container_from_image',
        return_value='mock-image-working-container-mock-1'
    )
    @patch.object(OpenSCAPGeneric, '_OpenSCAPGeneric__get_oscap_document_type')
    @patch('sh.buildah', create=True)
    def test_fail_downloading_open_scap_tailoring_file(
        self,
        buildah_mock,
        get_oscap_document_type_mock,
        create_container_from_image_mock,
        mount_container_mock,
        run_oscap_scan_mock,
        download_and_decompress_source_to_destination_mock
    ):
        oscap_document_type = 'Source Data Stream'
        oscap_eval_type = 'xccdf'
        oscap_input_definitions_uri = 'https://www.redhat.com/security/data/metrics/ds/v2/RHEL8/rhel-8.ds.xml.bz2'
        container_image_tag = 'localhost/mock-org/mock-image:v0.42.0-mock'
        oscap_tailoring_uri = 'https://raw.githubusercontent.com/ploigos/ploigos-example-oscap-content/main/xccdf_com.redhat.ploigos_profile_example_ubi8-tailoring-xccdf.xml'
        step_config = {
            'oscap-input-definitions-uri': oscap_input_definitions_uri,
            'oscap-profile': 'foo',
            'oscap-tailoring-uri': oscap_tailoring_uri,
            'container-image-tag': container_image_tag
        }
        oscap_eval_success = True
        oscap_eval_fails = None

        with TempDirectory() as temp_dir:
            # setup test
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            mount_path = '/does/not/matter/container-mount'

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='test',
                implementer='OpenSCAP',
                parent_work_dir_path=parent_work_dir_path
            )

            get_oscap_document_type_mock.return_value = oscap_document_type
            mount_container_mock.return_value = mount_path
            run_oscap_scan_mock.return_value = [
                oscap_eval_success,
                oscap_eval_fails
            ]

            # run test with mock error
            mock_error_msg = 'mock error downloading open scap file'
            download_and_decompress_source_to_destination_mock.side_effect = [
                "foo",
                RuntimeError(mock_error_msg)
            ]

            stdout_buff = StringIO()
            with redirect_stdout(stdout_buff):
                step_result = step_implementer._run_step()

            # verify results
            expected_step_result = StepResult(
                step_name='test',
                sub_step_name='OpenSCAP',
                sub_step_implementer_name='OpenSCAP'
            )
            expected_step_result.success = False
            expected_step_result.message = 'Error downloading OpenSCAP tailoring file: ' \
                f'{mock_error_msg}'
            self.assertEqual(expected_step_result, step_result)

            stdout = stdout_buff.getvalue()
            expected_container_name = 'mock-image-working-container-mock-1'
            self.assertRegex(
                stdout,
                re.compile(
                    rf".*Create container from image \({container_image_tag}\)"
                    rf".*Created container \({expected_container_name}\) from image \({container_image_tag}\)"
                    rf".*Mount container: {expected_container_name}"
                    rf".*Mounted container \({expected_container_name}\) with mount path: '{mount_path}'"
                    rf".*Download input definitions: {oscap_input_definitions_uri}"
                    rf".*Downloaded input definitions to: foo"
                    rf".*Download oscap tailoring file: {oscap_tailoring_uri}",
                    re.DOTALL
                )
            )



"""Constructs a given named StepImplementer using a given configuration, and runs it.
"""
import os
import fcntl

from ploigos_step_runner.config.config import Config
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.results import WorkflowResult
from ploigos_step_runner.step_implementer import StepImplementer
from ploigos_step_runner.utils.reflection import import_and_get_class


class StepRunner:
    """Enables the running of arbitrary Ploigos steps via StepImplementers.

    Parameters
    ----------
    config : Config, dict, list, str (file or directory)
        A Config object,
        or a dictionary that is a valid step runner configuration,
        or a string that is a path to a YAML or JSON file that is
        a valid step runner configuration,
        or a string that is a path to a directory containing one or more
        files that are valid YAML or JSON files that are valid
        configurations,
        or a list of any of the former.
    results_file_name : str, optional
        Path to the file for steps to write their results to
        Default: step-runner-results.yml
    work_dir_path : str, optional
        Path to the working folder for step_implementers for runtime files
        Default: step-runner-working

    Raises
    ------
    ValueError
        If given config is not of expected type.
    AssertionError
        If given config contains any invalid configurations.

    Returns
    -------
    Bool
        True if step completed successfully
        False if step returned an error message
    """

    __DEFAULT_MODULE = 'ploigos_step_runner.step_implementers'

    def __init__(
        self,
        config,
        results_file_name='step-runner-results.yml',
        work_dir_path='step-runner-working'
    ):
        if isinstance(config, Config):
            self.__config = config
        else:
            self.__config = Config(config)

        self.__results_file_name = results_file_name
        self.__work_dir_path = work_dir_path

        self.__workflow_result = None

    @property
    def config(self):
        """
        Returns
        -------
        Config
            Configuration used by this factory.
        """
        return self.__config

    @property
    def results_file_path(self):
        """Get the full path to the results file.

        Returns
        -------
        str
            Full path to the results file.
        """
        return os.path.join(self.__work_dir_path, self.__results_file_name)

    @property
    def workflow_result_pickle_file_path(self):
        """
        Get the full path to the workflow result pickle file.
        (The 'pickle' file contains the serialized list of step results.)
        The name of the pickle file is the basename of the results_file_name.

        Returns
        -------
        str
           Full path to the workflow pickle (serialized) file.
        """
        pickle_filename = os.path.splitext(self.__results_file_name)[0] + '.pkl'
        return os.path.join(self.__work_dir_path, pickle_filename)

    @property
    def workflow_result(self):
        """
        Returns
        -------
        WorkflowResult
            Object containing a list of dictionary of step results
            from previous steps.
        """
        if not self.__workflow_result:
            self.__workflow_result = WorkflowResult.load_from_pickle_file(
                pickle_filename=self.workflow_result_pickle_file_path
            )
        return self.__workflow_result

    def run_step(self, step_name, environment=None):
        """
        Call the given step.

        Parameters
        ----------
        step_name : str
            Ploigos step to run.
        environment : str, optional
            Name of the environment the step is being run in. Used to determine environment
            specific global defaults and step configuration.

        Raises
        ------
        # todo: doc needs to be updated
        StepRunnerException
            If no StepImplementers have been registered for the given step_name
            If no specific StepImplementer name specified in sub step config
                and no default StepImplementer registered for given step_name.
            If no StepImplementer registered for given step with given implementer name.
        Returns
        -------
        Bool
           True if step completed successfully
           False if step returned an error message
        """

        sub_step_configs = self.config.get_sub_step_configs(step_name)
        assert len(sub_step_configs) != 0, \
            f"Can not run step ({step_name}) because no step configuration provided."

        # for each sub step in the step config get the step implementer and run it
        aggregate_success = True
        for sub_step_config in sub_step_configs:
            sub_step_implementer_name = sub_step_config.sub_step_implementer_name

            step_implementer_class = StepRunner.__get_step_implementer_class(
                step_name,
                sub_step_implementer_name)

            # create the StepImplementer instance
            sub_step = step_implementer_class(
                parent_work_dir_path=self.__work_dir_path,
                config=sub_step_config,
                environment=environment,
                workflow_result=self.workflow_result
            )

            # run the step
            step_result = sub_step.run_step()

            # save the step results
            self.workflow_result.add_step_result(
                step_result=step_result
            )

            # acquire the pickle lock
            with open(
                    self.workflow_result_pickle_file_path + '.lock',
                    'w',
                    encoding='utf-8'
            ) as pickle_lock:
                fcntl.flock(pickle_lock, fcntl.LOCK_EX)

                # we are locked
                try:
                    self.workflow_result.merge_with_pickle_file(
                        pickle_filename=self.workflow_result_pickle_file_path
                    )
                    self.workflow_result.write_to_pickle_file(
                        pickle_filename=self.workflow_result_pickle_file_path
                    )
                    self.workflow_result.write_results_to_yml_file(
                        yml_filename=self.results_file_path
                    )
                finally:
                    fcntl.flock(pickle_lock, fcntl.LOCK_UN)

            # aggregate success
            aggregate_success = (aggregate_success and step_result.success)

            # if this sub step failed and not configured to continue on failure, bail
            # else execute next sub step and continue aggregating success
            if (not step_result.success) and \
                    (not sub_step_config.sub_step_contine_sub_steps_on_failure):
                break

        return aggregate_success

    @staticmethod
    def __get_step_implementer_class(step_name, step_implementer_name):
        """Given a step name and a step implementer name dynamically loads the Class.

        Parameters
        ----------
        step_name : str
            Name of the step to load the given step implementer for.
            This is only used if the given step_implementer_name does not include
            a module path.
        step_implementer_name : str
            Either the short name of a StepImplementer class which will be dynamically
            loaded from the 'ploigos_step_runner.step_implementers.{step_name}' module or
            A class name that includes a dot seperated module name to load the Class from.

        Returns
        -------
        StepImplementer
            Dynamically loaded subclass of StepImplementer for given step name with
            given step implementer name.

        Raises
        ------
        StepRunnerException
            If could not find class to load
            If loaded class is not a subclass of StepImplementer
        """
        parts = step_implementer_name.split('.')
        class_name = parts.pop()
        module_name = '.'.join(parts)

        if not module_name:
            step_module_part = step_name.replace('-', '_')
            module_name = f"{StepRunner.__DEFAULT_MODULE}.{step_module_part}"

        clazz = import_and_get_class(module_name, class_name)
        if not clazz:
            raise StepRunnerException(
                f"Could not dynamically load step ({step_name}) step implementer" +
                f" ({step_implementer_name}) from module ({module_name})" +
                f" with class name ({class_name})"
            )
        if not issubclass(clazz, StepImplementer):
            raise StepRunnerException(
                f"Step ({step_name}) is configured to use step implementer" +
                f" ({step_implementer_name}) from module ({module_name}) with" +
                f" class name ({class_name}), and dynamically loads as class ({clazz})" +
                f" which is not a subclass of required parent class ({StepImplementer}).")

        return clazz



"""`StepImplementers` for the `generate_evidence` step.
"""

from ploigos_step_runner.step_implementers.generate_evidence.generate_evidence import \
    GenerateEvidence
from ploigos_step_runner.step_implementers.generate_evidence.rekor_sign_evidence import \
    RekorSignEvidence



import os
import re

import tests.step_implementers.shared.test_openscap_generic
from ploigos_step_runner.step_implementers.container_image_static_compliance_scan import \
    OpenSCAP
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase


class BaseTestStepImplementerContainerImageStaticComplianceScanOpenSCAP(BaseStepImplementerTestCase):
    def create_step_implementer(
            self,
            step_config={},
            step_name='test',
            implementer='OpenSCAP',
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=OpenSCAP,
            step_config=step_config,
            step_name=step_name,
            implementer=implementer,
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

class TestStepImplementerContainerImageStaticComplianceScanOpenSCAP_step_implementer_config_defaults(
    BaseTestStepImplementerContainerImageStaticComplianceScanOpenSCAP,
    tests.step_implementers.shared.test_openscap_generic.TestStepImplementerSharedOpenSCAPGeneric_step_implementer_config_defaults
):
    pass

class TestStepImplementerContainerImageStaticComplianceScanOpenSCAP__required_config_or_result_keys(
    BaseTestStepImplementerContainerImageStaticComplianceScanOpenSCAP
):
    def test_result(self):
        required_keys = OpenSCAP._required_config_or_result_keys()
        expected_required_keys = [
            'oscap-profile',
            'oscap-input-definitions-uri',
            'container-image-tag',
            'container-image-pull-repository-type',
            ['container-image-pull-repository-type', 'container-image-repository-type']
        ]
        self.assertEqual(required_keys, expected_required_keys)

class TestStepImplementerContainerImageStaticVulnerabilityScanOpenSCAP__validate_required_config_or_previous_step_result_artifact_keys(
    BaseTestStepImplementerContainerImageStaticComplianceScanOpenSCAP,
    tests.step_implementers.shared.test_openscap_generic.TestStepImplementerSharedOpenSCAPGeneric__validate_required_config_or_previous_step_result_artifact_keys
):
    pass

class TestStepImplementerContainerImageStaticComplianceScanOpenSCAP___get_oscap_document_type(
    BaseTestStepImplementerContainerImageStaticComplianceScanOpenSCAP,
    tests.step_implementers.shared.test_openscap_generic.TestStepImplementerSharedOpenSCAPGeneric___get_oscap_document_type
):
    pass

class TestStepImplementerContainerImageStaticComplianceScanOpenSCAP___get_oscap_eval_type_based_on_document_type(
    BaseTestStepImplementerContainerImageStaticComplianceScanOpenSCAP,
    tests.step_implementers.shared.test_openscap_generic.TestStepImplementerSharedOpenSCAPGeneric___get_oscap_eval_type_based_on_document_type
):
    pass

class TestStepImplementerContainerImageStaticComplianceScanOpenSCAP___run_oscap_scan(
    BaseTestStepImplementerContainerImageStaticComplianceScanOpenSCAP,
    tests.step_implementers.shared.test_openscap_generic.TestStepImplementerSharedOpenSCAPGeneric___run_oscap_scan
):
    pass

class TestStepImplementerContainerImageStaticComplianceScanOpenSCAP__run_step(
    BaseTestStepImplementerContainerImageStaticComplianceScanOpenSCAP,
    tests.step_implementers.shared.test_openscap_generic.TestStepImplementerSharedOpenSCAPGeneric__run_step
):
    pass



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring
import os

from git import Repo
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase
from tests.helpers.test_utils import create_git_commit_with_sample_file
from ploigos_step_runner import StepResult
from ploigos_step_runner.step_implementers.generate_metadata import Git


class TestStepImplementerGitGenerateMetadata(BaseStepImplementerTestCase):
    def create_step_implementer(
            self,
            step_config={},
            step_name='',
            implementer='',
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=Git,
            step_config=step_config,
            step_name=step_name,
            implementer=implementer,
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

    def test_step_implementer_config_defaults(self):
        defaults = Git.step_implementer_config_defaults()
        expected_defaults = {
            'repo-root': './',
            'build-string-length': 7
        }
        self.assertEqual(defaults, expected_defaults)

    def test__required_config_or_result_keys(self):
        required_keys = Git._required_config_or_result_keys()
        expected_required_keys = ['repo-root', 'build-string-length']
        self.assertEqual(required_keys, expected_required_keys)

    def test_run_step_pass(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            repo = Repo.init(str(temp_dir.path))

            create_git_commit_with_sample_file(temp_dir, repo)

            step_config = {
                'repo-root': str(temp_dir.path)
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='generate-metadata',
                implementer='Git',
                parent_work_dir_path=parent_work_dir_path,
            )

            result = step_implementer._run_step()

            # cheating because we don't want to fully mock this yet
            self.assertTrue(result.success, True)

    def test_root_dir_is_not_git_repo(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                'repo-root': '/'
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='generate-metadata',
                implementer='Git',
                parent_work_dir_path=parent_work_dir_path,
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='generate-metadata',
                sub_step_name='Git',
                sub_step_implementer_name='Git'
            )
            expected_step_result.success = False
            expected_step_result.message = 'Given directory (repo_root) is not a Git repository'

            self.assertEqual(result, expected_step_result)

    def test_root_dir_is_bare_git_repo(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            Repo.init(str(temp_dir.path), bare=True)

            step_config = {
                'repo-root': str(temp_dir.path)
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='generate-metadata',
                implementer='Git',
                parent_work_dir_path=parent_work_dir_path,
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='generate-metadata',
                sub_step_name='Git',
                sub_step_implementer_name='Git'
            )
            expected_step_result.success = False
            expected_step_result.message = 'Given directory (repo_root) is a bare Git repository'

            self.assertEqual(result, expected_step_result)

    def test_no_commit_history(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            Repo.init(str(temp_dir.path))

            step_config = {
                'repo-root': str(temp_dir.path)
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='generate-metadata',
                implementer='Git',
                parent_work_dir_path=parent_work_dir_path,
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(
                step_name='generate-metadata',
                sub_step_name='Git',
                sub_step_implementer_name='Git'
            )
            expected_step_result.add_artifact(
                name='pre-release',
                value='master'
            )
            expected_step_result.success = False
            expected_step_result.message = 'Given directory (repo_root) is a ' \
                                           'git branch (git_branch) with no commit history'

            self.assertEqual(result, expected_step_result)

    def test_git_repo_with_single_commit_on_master(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            repo = Repo.init(str(temp_dir.path))

            create_git_commit_with_sample_file(temp_dir, repo)

            step_config = {
                'repo-root': str(temp_dir.path)
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='generate-metadata',
                implementer='Git',
                parent_work_dir_path=parent_work_dir_path,
            )

            result = step_implementer._run_step()

            # cheating because we don't want to fully mock this yet
            self.assertTrue(result.success, True)

    def test_git_repo_with_single_commit_on_feature(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            repo = Repo.init(str(temp_dir.path))

            create_git_commit_with_sample_file(temp_dir, repo)

            # checkout a feature branch
            git_new_branch = repo.create_head('feature/test0')
            git_new_branch.checkout()

            step_config = {
                'repo-root': str(temp_dir.path)
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='generate-metadata',
                implementer='Git',
                parent_work_dir_path=parent_work_dir_path,
            )

            result = step_implementer._run_step()

            # cheating because we don't want to fully mock this yet
            self.assertEqual(result.get_artifact_value('pre-release'), 'feature_test0')
            self.assertTrue(result.success, True)

    def test_directory_is_detached(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            repo = Repo.init(str(temp_dir.path))

            # create commits
            create_git_commit_with_sample_file(temp_dir, repo, 'test0')
            create_git_commit_with_sample_file(temp_dir, repo, 'test1')

            # detach head
            repo.git.checkout('master^')

            step_config = {
                'repo-root': str(temp_dir.path)
            }

            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='generate-metadata',
                implementer='Git',
                parent_work_dir_path=parent_work_dir_path,
            )

            result = step_implementer._run_step()

            expected_step_result = StepResult(step_name='generate-metadata', sub_step_name='Git',
                                              sub_step_implementer_name='Git')
            expected_step_result.success = False
            expected_step_result.message = 'Expected a Git branch in given directory (repo_root) ' \
                                           'but has a detached head'

            self.assertEqual(result, expected_step_result)



"""`StepImplementers` for the `report` step.
"""

from ploigos_step_runner.step_implementers.report.result_artifacts_archive import \
    ResultArtifactsArchive
from ploigos_step_runner.step_implementers.report.rekor_sign_report import \
    RekorSignReport



"""`StepImplementer` for the `automated-governance` step using Rekor.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:

  * static configuration
  * runtime configuration
  * previous step results

Configuration Key       | Required? | Default | Description
-------------------     |-----------|---------|-----------
`rekor-server-url`      | Yes       |         | URL for Rekor server to upload artifact(s) to.
`signer-pgp-private-key`| Yes       |         | PGP Private Key used to sign the image

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key | Description
--------------------|------------
`rekor-uuid`        | UUID of Rekor entry created by artifact upload.
`rekor-entry`       | DEPRECIATED. Will likely be removed in next release.
"""

from base64 import b64encode
import json
import sys
from io import StringIO
import sh
from ploigos_step_runner import StepImplementer, StepResult
from ploigos_step_runner.utils.file import base64_encode, get_file_hash,\
    download_source_to_destination
from ploigos_step_runner.utils.io import \
    create_sh_redirect_to_multiple_streams_fn_callback
from ploigos_step_runner.utils.pgp import detach_sign_with_pgp_key
from ploigos_step_runner.utils.pgp import import_pgp_key
from ploigos_step_runner.utils.pgp import export_pgp_public_key


DEFAULT_CONFIG = {}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'rekor-server-url',
    'signer-pgp-private-key'
]
class RekorSignGeneric(StepImplementer):  # pylint: disable=too-few-public-methods
    """`StepImplementer` for the generic Rekor class.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        workflow_result,
        parent_work_dir_path,
        config,
        artifact_to_sign_uri_config_key,
        environment=None
    ):
        self.__artifact_to_sign_uri_config_key = artifact_to_sign_uri_config_key
        super().__init__(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=environment
        )

    @property
    def artifact_to_sign_uri_config_key(self):
        """Getter for the artifact to sign config key.

        Returns
        -------
        str
            Config key that is used to obtain the artifact
            that needs to be signed
        """
        return self.__artifact_to_sign_uri_config_key

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.

        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _create_rekor_entry( # pylint: disable=no-self-use
        self,
        signer_pgp_public_key,
        signer_pgp_private_key_fingerprint,
        path_to_file,
        artifact_to_sign_uri
    ):
        """Method to generate a rekor entry

        Parameters
        ----------
        signer_pgp_public_key: str
            Public key obtained from the private key fingerprint
        signer_pgp_private_key_fingerprint: str
            PGP fingerprint obtained from importing the private key
        path_to_file: str
            Path to file to be signed
        artifact_to_sign_uri: str
            URI where artifact was pulled from

        Returns
        -------
        dict
            Dictionary containing the rekor generated entry.


        """

        file_hash = get_file_hash(path_to_file)
        sig_file = path_to_file + '.asc'

        base64_public_key = b64encode(signer_pgp_public_key.encode('utf-8')).decode('utf-8')

        detach_sign_with_pgp_key(
            output_signature_path=sig_file,
            file_to_sign_path=path_to_file,
            pgp_private_key_fingerprint=signer_pgp_private_key_fingerprint
        )

        #Base 64 encode the file

        with open(path_to_file, 'rb') as file_bytes:
            base64_encoded_data = b64encode(file_bytes.read()).decode('utf-8')

        rekor_entry = {
            "kind": "rekord",
            "apiVersion": "0.0.1",
            "spec": {
                "signature": {
                    "format": "pgp",
                    "content": base64_encode(sig_file),
                    "publicKey": {
                        "content": base64_public_key
                    }
                },
                "data": {
                    "content": base64_encoded_data,
                    "hash": {
                        "algorithm": "sha256",
                        "value": file_hash
                    }
                },
                "extraData": {
                    "signed-artifact-uri": artifact_to_sign_uri

                }
            }
        }
        return rekor_entry

    def _upload_to_rekor(
        self,
        rekor_server,
        rekor_entry
    ):
        """Method to upload a rekor entry to provided rekor server

        Parameters
        ----------
        rekor_server: str
            URL to rekor server
        signer_pgp_private_key_fingerprint: str
            PGP fingerprint obtained from importing the private key
        path_to_file: str
            Path to file to be signed
        artifact_to_sign_uri: str
            URI where artifact was pulled from

        Returns
        -------
        str:
            Returns rekor uuid returned from upload command

        """

        rekor_entry_path = self.write_working_file(
            filename='entry.json',
            contents=bytes(json.dumps(rekor_entry), 'utf-8')
        )
        rekor_upload_stdout_result = StringIO()
        rekor_upload_stdout_callback = create_sh_redirect_to_multiple_streams_fn_callback([
            sys.stdout,
            rekor_upload_stdout_result
        ])
        rekor = sh.rekor( # pylint: disable=no-member
            'upload',
            '--rekor_server',
            rekor_server,
            '--entry',
            rekor_entry_path,
            _out=rekor_upload_stdout_callback,
            _err_to_out=True,
            _tee='out'
        )
        rekor_uuid = str(rekor).rsplit('/', maxsplit=1)[-1].strip(' \n')
        return rekor_uuid

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)
        rekor_server = self.get_value('rekor-server-url')

        work_dir = self.work_dir_path
        artifact_to_sign_uri = self.get_value(self.artifact_to_sign_uri_config_key)
        #Download artifact that needs to be signed and place at work_dir.
        #Path to file is returned as string
        path_to_file = download_source_to_destination(
            artifact_to_sign_uri,
            work_dir)

        # get the pgp private key to sign the image with
        signer_pgp_private_key = self.get_value(
            'signer-pgp-private-key'
        )

        # import the PGP key and get the finger print
        signer_pgp_private_key_fingerprint = import_pgp_key(
            pgp_private_key=signer_pgp_private_key
        )

        signer_pgp_public_key = export_pgp_public_key(signer_pgp_private_key_fingerprint)

        rekor_entry = self._create_rekor_entry(
            signer_pgp_public_key,
            signer_pgp_private_key_fingerprint,
            path_to_file,
            artifact_to_sign_uri
        )

        rekor_uuid = self._upload_to_rekor(
            rekor_server,
            rekor_entry
        )
        step_result.add_artifact(
                name='rekor-entry',
                value=rekor_entry
        )
        step_result.add_artifact(
                name='rekor-uuid',
                value=rekor_uuid
        )
        rekor_uri = rekor_server + '/api/v1/log/entries/' + rekor_uuid
        step_result.add_artifact(
                name='rekor-entry-uri',
                value=rekor_uri
        )
        return step_result



"""`StepImplementers` for the `push-artifacts` step.
"""

from ploigos_step_runner.step_implementers.push_artifacts.maven_deploy import MavenDeploy
from ploigos_step_runner.step_implementers.push_artifacts.maven import Maven




import os
from ploigos_step_runner.step_implementers.report.rekor_sign_report import RekorSignReport
from testfixtures import TempDirectory
from tests.step_implementers.shared.test_rekor_sign_generic import \
    TestStepImplementerSharedRekorSignGeneric

class TestStepImplementerRekorReport(TestStepImplementerSharedRekorSignGeneric):
    def create_step_implementer_rekor_evidence(
            self,
            step_config={},
            workflow_result=None,
            parent_work_dir_path='',
    ):

        step_implementer = self.create_given_step_implementer(
            step_implementer=RekorSignReport,
            step_config=step_config,
            step_name='report',
            implementer='RekorSignReport',
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
            )
        return step_implementer

    def test__validate_rekor_constructor(self):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')
            step_config = {
                    'rekor-server-url': self.TEST_REKOR_SERVER_URL,
                    'signer-pgp-private-key': self.TEST_PGP_SIGNER_PRIVATE_KEY
                }
            step_implementer = self.create_step_implementer_rekor_evidence(
                    step_config=step_config,
                    parent_work_dir_path=parent_work_dir_path
                )

            artifact_to_sign_uri_config_key = step_implementer.artifact_to_sign_uri_config_key

            self.assertEqual(artifact_to_sign_uri_config_key, 'results-archive-uri')

